from fnmatch import fnmatch
try:
    import configparser as configparser
except ImportError:
    import configparser
from snaprotator.logsetup import UserLogger
from snaprotator.errors import EXIT_CODE_CONFIG_FILE_NOT_FOUND
from snaprotator.utils import unmask
import sys

from marshmallow import Schema, ValidationError, validates, validates_schema, pre_load, post_dump, validate
from marshmallow.fields import String, Integer, Boolean, Nested, Email, Dict, List


log = UserLogger(__name__)


class HiddenString(str):
    def __repr__(self):
        # replacing the "repr" will hide the password from interactive prompts, tracebacks, etc.
        return "*" * 8

    def __str__(self):
        # replacing the "str" will hide the password from interactive prompts, tracebacks, etc.
        return "*" * 8


def load_config(arguments):
    config = configparser.ConfigParser(allow_no_value=True)
    configpath = arguments['--configfile']

    try:
        with open(configpath) as f:
            config.readfp(f)
    except IOError:
        sys.stderr.write(
            'Config file path not found. Exiting. ({})\n'.format(configpath))
        log.uabort('Config file path not found. ({})', configpath,
                   exit_code=EXIT_CODE_CONFIG_FILE_NOT_FOUND)

    config_dict = {}
    for section in config.sections():
        config_dict[section] = dict((key, True if value is None else value)
                                    for key, value in config.items(section))
    return config_dict


def update_arguments_to_config(arguments, config):
    def key_set(section, key, value):
        s = config.get(section, dict())
        config[section] = s
        s[key] = value
    lookup = {
        "--hostname": ('infinibox', 'host'),
        "--retain": ('global', 'snaps_to_keep'),
        "--lockdays": ('global', 'snap_lock_days'),
        "--maxlockdays": ('global', 'max_snap_lock_days'),
        "--basename": ('global', 'snap_basename'),
        "--debug": ('logging', 'debug'),
        "--logfile": ('logging', 'logpath'),
        "--norotation": ('logging', 'rotation'),
        "--password": ('infinibox', 'password'),
        "--quiet": ('logging', 'quiet'),
        "--syslog": ('logging', 'syslog'),
        "--user": ('infinibox', 'username'),
        "--volumes": ('infinibox', 'volumes'),
        "--purge": ('infinibox', 'purge'),
        "--dryrun": ('global', 'dryrun'),
        "--force_report": ('global', 'force_report'),
        "--delete": ('global', 'delete'),
        "--list": ('global', 'list'),
        "--listjson": ('global', 'listjson')
    }
    for arg_key, (section, key) in list(lookup.items()):
        if arg_key in arguments and arguments[arg_key] is not None and arguments[arg_key] is not False:
            key_set(section, key, arguments[arg_key])


def validate_config(config):
    class Global(Schema):
        snap_basename = String(required=True)
        dryrun = Boolean()
        delete = Boolean()
        _list = Boolean(data_key='list', attribute='list')
        listjson = Boolean()
        force_report = Boolean(missing=False)
        use_volid = Boolean(missing=True)
        sa_internal = Boolean(missing=False)
        allow_replicas = Boolean(missing=False)
        snap_mapped_fs_snapshots = Boolean(missing=False)
        snap_both_aa_peers = Boolean(missing=False)
        single_instance = Boolean(missing=False)
        max_snap_lock_days = Integer(
            validate=validate.Range(min=1, max=90), missing=30)
        snap_lock_days = Integer(validate=validate.Range(min=1, max=90))
        snaps_to_keep = Integer(validate=validate.Range(min=1, max=998))

        @validates_schema
        def _dependencies(self, data, skip_on_field_errors=True, **kwargs):
            if sum([1 for a in [data.get('delete'), data.get('listjson'), data.get('list')] if a]) > 1:
                raise ValidationError(
                    {'list': "Only one of list/listjson/delete should be specified"})

            if data.get('snap_lock_days', 0) > data.get('max_snap_lock_days', 30):
                raise ValidationError(
                    {'snap_lock_days': "snap_lock_days can't be more than max_snap_lock_days"})

    class Infinibox(Schema):
        host = String(required=True)
        password = String(required=True)
        username = String(required=True)
        purge = Boolean()
        volumes = String()

        @validates('password')
        def _passwd(self, value, **kwargs):
            try:
                unmask(value)
            except Exception as e:
                raise ValidationError(str(e))

        @post_dump
        def _mask(self, data, **kwargs):
            data['password'] = HiddenString(unmask(data['password']))
            return data

    class Logging(Schema):
        logpath = String()
        emailuser = Email()
        emailpassword = String()
        emailsender = Email()
        emailrcpt = List(Email)
        emailsubject = String(missing="")
        emailserver = String()
        emailport = Integer(validate=validate.Range(
            min=1, max=56535), missing=25)
        quiet = Boolean()
        backup_radar_email_format = Boolean(missing=False)
        email_skip_success = Boolean(missing=False)
        logs_to_keep = Integer(validate=validate.Range(min=1, max=100))
        rotation = Boolean(missing=True)
        debug = Boolean(missing=False)
        email = Boolean(missing=False)
        suppress = Boolean(missing=True)
        syslog = Boolean(missing=False)

        @validates('emailpassword')
        def _passwd(self, value, **kwargs):
            try:
                unmask(value)
            except Exception as e:
                raise ValidationError(str(e))

        @post_dump
        def _mask(self, data, **kwargs):
            if data.get('emailpassword'):
                data['emailpassword'] = HiddenString(
                    unmask(data['emailpassword']))
            return data

        @validates_schema
        def _dependencies(self, data, skip_on_field_errors=True, **kwargs):
            if data.get('email') and not all([data.get('emailsender'), data.get('emailrcpt'), data.get('emailserver')]):
                raise ValidationError(
                    {'email': 'emailsender,emailrcpt,emailserver are mandatory when email is enabled'})

            if data.get('email') and not (all([data.get('emailpassword'), data.get('emailuser')]) or not any([data.get('emailpassword'), data.get('emailuser')])):
                raise ValidationError(
                    {'email': 'emailuser and emailpassword shoud both be present or absent'})

            if data.get('logs_to_keep') and not data.get('rotation'):
                raise ValidationError(
                    {'logs_to_keep': 'logs_to_keep requires rotation to be set'})

        @pre_load
        def _emailrcpt(self, data, **kwargs):
            emailrcpt = data.get('emailrcpt')
            if emailrcpt:
                emailrcpts = [a.strip() for a in emailrcpt.split(',')]
                data['emailrcpt'] = emailrcpts
            return data
    class Vcenter(Schema):
        host = String(required=True)
        password = String(required=True)
        username = String(required=True)
        parallel = Integer(validate=validate.Range(min=1, max=20))
        consistent_only = Boolean()
        include_templates = Boolean()

        @validates('password')
        def _passwd(self, value, **kwargs):
            try:
                unmask(value)
            except Exception as e:
                raise ValidationError(str(e))

        @post_dump
        def _mask(self, data, **kwargs):
            data['password'] = HiddenString(unmask(data['password']))
            return data

    class Config(Schema):
        _global = Nested(Global, required=True,
                         data_key='global', attribute='global')
        infinibox = Nested(Infinibox, required=True)
        logging = Nested(Logging)
        vcenter = Nested(Vcenter)
        datastores = Dict()
        nomemory = Dict()
        noquiesce = Dict()
        vms = Dict()
        volumes = Dict()
        fs = Dict()

        @validates_schema
        def _dependencies(self, data, skip_on_field_errors=True, **kwargs):
            if all([data.get('infinibox', {}).get('purge'), data.get('global', {}).get('snaps_to_keep')]):
                raise ValidationError(
                    {'global': "purge and snaps_to_keep can't be specified together"})

            if not any([
                data.get('global', {}).get('snaps_to_keep'),
                data.get('infinibox', {}).get('purge'),
                data.get('global', {}).get('list'),
                data.get('global', {}).get('listjson'),
                data.get('global', {}).get('delete'),
            ]):
                raise ValidationError(
                    {'global': "One of purge/snaps_to_keep/list/listjson/delete should be specified"})

            if sum([1 for a in [
                data.get('infinibox', {}).get('purge'),
                data.get('global', {}).get('list'),
                data.get('global', {}).get('listjson'),
                data.get('global', {}).get('delete'),
            ] if a]) > 1:
                raise ValidationError(
                    {'global': 'Only one of purge/list/listjson/delete can be specified'})

            if not any([
                data.get('infinibox', {}).get('purge'),
                data.get('global', {}).get('list'),
                data.get('global', {}).get('listjson'),
                data.get('global', {}).get('delete'),
            ]):
                if all([data.get('global', {}).get('snaps_to_keep'), data.get('infinibox', {}).get('purge')]):
                    raise ValidationError(
                        {'purge': "Either --purge or --snaps_to_keep should be specified, not both."})

                if not any(['volumes' in data.get('infinibox'), 'volumes' in data, 'fs' in data, 'vms' in data]):
                    raise ValidationError(
                        {'volumes': 'Either ([vcenter] section or [volumes] or [fs] or --volumes should be configured.'})

                if any(['volumes' in data.get('infinibox'), 'volumes' in data, 'fs' in data]) and 'vcenter' in data:
                    raise ValidationError(
                        {'volumes': "[vcenter] section and --volumes/[volumes]/[fs] can't be used simultaneously."})

                if data.get('vcenter') and not all(['vms' in data, 'datastores' in data, 'noquiesce' in data, 'nomemory' in data]):
                    raise ValidationError(
                        {'vcenter': "fields 'vms' ,'noquiesce','nomemory' and 'datastores' are required"})

                if any(['vms' in data, 'datastores' in data, 'noquiesce' in data, 'nomemory' in data]) and not data.get('vcenter'):
                    raise ValidationError(
                        {'vcenter': "[vcenter] is required with fields 'vms' ,'noquiesce','nomemory' and 'datastores'"})

                if all(['volumes' in data.get('infinibox'), any(['volumes' in data, 'fs' in data])]):
                    raise ValidationError(
                        {'volumes': "--volumes([infinibox]volumes) section and [volumes]/[fs] can\'t be used simultaneously."})

            if data.get('infinibox', {}).get('purge'):
                data['global']['snaps_to_keep'] = 1

    try:
        a = Config().load(config)
        cfg = Config().dump(a)
        return True, cfg
    except ValidationError as e:
        return False, str(e)


def read_patterns_from_config(config, entity):
    return config.get(entity, [])


def is_use_volid(config):
    return 'global' in config and 'use_volid' in config['global'] and config['global']['use_volid']
