from distutils.version import LooseVersion
from snaprotator.logsetup import UserLogger, has_warnings
from infinisdk import InfiniBox
from infinisdk.core.exceptions import APICommandFailed, ChangedDuringIteration
from contextlib import contextmanager
from snaprotator import errors
import json
from time import time, gmtime
import pickle
from snaprotator.utils import dummy_context_mgr
from time import sleep
from fnmatch import fnmatch
from re import match
from datetime import timedelta, datetime
from snaprotator.config import read_patterns_from_config
from snaprotator.utils import timed
from snaprotator.rotation import trim_number_of_snapshots, create_latest_snapshot, filter_irrelevant_snapshots, IMMUTABLE_SUPPORTED_VER
from snaprotator.__version__ import __version__ as snaprotator_version

RETRY_ON_CHANGED_DURING_ITERATION_COUNT = 3

MAX_SYSTEM_SNAPSHOTS = 100000
MAX_VOL_SNAPSHOTS = 998
MAX_FS_SNAPSHOTS = 9998
MAX_CG_SNAPSHOTS = 248
MAX_CG_SNAPSHOTS_V6 = 998


MIN_SUPPORTED_VER = LooseVersion("1.5.0")
MAX_SUPPORTED_VER = LooseVersion("8.0.1")
CG_SUPPORTED_VER = LooseVersion("2.2.0")
FS_SUPPORTED_VER = LooseVersion("2.1.0")

log = UserLogger(__name__)


def _retry_on_changed_during_iteration(f):
    def wrapper(*args, **kwargs):
        for i in range(RETRY_ON_CHANGED_DURING_ITERATION_COUNT + 1):
            try:
                return f(*args, **kwargs)
            except ChangedDuringIteration:
                if i >= RETRY_ON_CHANGED_DURING_ITERATION_COUNT:
                    raise
                log.uinfo(
                    "Objects changed during processing, retrying operation. Please ignore any warning about queried path size changed during iteration.")
                sleep(0.5)
    return wrapper


@contextmanager
def infinibox_api_error_wrapper(hostname, abort=True):
    if abort:
        log_handler = log.uabort
        log_handler_exc = log.uabort_with_exc_info
    else:
        log_handler = log_handler_exc = log.uwarn
    try:
        yield
    except APICommandFailed as e:
        if e.error_code in ('AUTHENTICATION_REQUIRED', 'UNAUTHORIZED'):
            log_handler("Could not connect to {} API: username/password error.".format(hostname),
                        exit_code=errors.EXIT_CODE_IBOX_LOGIN_FAILED)
        else:
            log_handler_exc(
                "Caught APICommandFailed", exit_code=errors.EXIT_CODE_IBOX_API_FAILED)
    except SystemExit:
        raise
    except Exception as e:
        message = getattr(e, 'message', 'Failed to connect, aborting.')
        log_handler_exc(
            message, exit_code=errors.EXIT_CODE_IBOX_NETWORK_ERROR)


def load_cookie(infinibox, username, hostname):
    try:
        with open('/tmp/snaprotator_cookie_{}_{}'.format(username, hostname), 'rb') as f:
            return pickle.load(f)
    except Exception:
        log.debug("Cookie not found.")
    return None


def save_cookie(infinibox, username, hostname):
    try:
        with open('/tmp/snaprotator_cookie_{}_{}'.format(username, hostname), 'wb') as f:
            pickle.dump(infinibox.api._session.cookies.copy(), f)
    except Exception as e:
        log.debug("Failed to save cookie")
        log.debug(repr(e))


def my_login(infinibox, username, hostname):
    cookie = load_cookie(infinibox, username, hostname)
    if not cookie:
        log.debug("Cookie not found. Logging in.")
        infinibox.login()
        cookie = save_cookie(infinibox, username, hostname)
    else:
        log.debug("Loaded saved cookie")
        infinibox.api._session.cookies = cookie
        try:
            infinibox.get_version()
            # It didn't use the cookie, but looged in
            if infinibox.is_logged_in():
                log.debug("Logged in, cookie expired. Saving.")
                save_cookie(infinibox, username, hostname)
        except Exception:
            log.debug("Cookie not accepted")
            infinibox.api._session.cookies.clear()
            infinibox.login()
            save_cookie(infinibox, username, hostname)


def _infinibox_object(hostname, username, password, ssl=True):
    return InfiniBox(hostname, use_ssl=ssl, auth=(username, password))


def connect_to_infinibox(config, ssl=True, abort=True):
    hostname = config['infinibox']['host']
    username = config['infinibox']['username']
    password = config['infinibox']['password']

    log.uinfo("Connecting to the Infinibox '{}'".format(hostname))
    with infinibox_api_error_wrapper(hostname, abort):
        infinibox = _infinibox_object(hostname, username, password, ssl)
        my_login(infinibox, username, hostname)
        version = infinibox.get_version()
        log.info("Infinibox version: {}".format(version))
    return infinibox


def report_metadata_stats(infinibox, prefix, sa_internal=False):
    try:
        runs, fs_success, fs_failure, vol_success, vol_failure, warnings = calculate_metadata_stats(
            infinibox, prefix)
        log.info(
            'Total runs for policy "{}" in the last 24 hours: {}, runs with warnings: {}', prefix, runs, warnings)
        log.info('FS success: {} FS failure: {} VOL/CG Success: {} VOL/CG failure: {}',
                 fs_success, fs_failure, vol_success, vol_failure)
        event_24_h(infinibox, runs, prefix, fs_success,
                   fs_failure, vol_success, vol_failure, warnings, sa_internal)
    except Exception as e:
        log.uwarn("Failed to report metadata stats. {}".format(repr(e)))


def calculate_metadata_stats(infinibox, prefix):
    pool = find_smallest_pool(infinibox)
    if not pool:
        log.debug("Couldn't find pool - not caclulating stats")
        return 0, 0, 0, 0, 0, 0
    else:
        metadata = get_current_metadata(pool, prefix)
        keys = [x for x in metadata]
        fs_success = 0
        fs_failure = 0
        vol_success = 0
        vol_failure = 0
        warnings = 0
        for entry in keys:
            fs_success += metadata[entry].get('data', {}).get(
                'fs_success', metadata[entry].get('d', {}).get('fs', 0))
            fs_failure += metadata[entry].get('data', {}).get(
                'fs_failure', metadata[entry].get('d', {}).get('ff', 0))
            vol_success += metadata[entry].get('data', {}).get(
                'vol_success', metadata[entry].get('d', {}).get('vs', 0))
            vol_failure += metadata[entry].get('data', {}).get(
                'vol_failure', metadata[entry].get('d', {}).get('vf', 0))
            warnings += metadata[entry].get('data', {}).get(
                'warnings', metadata[entry].get('d', {}).get('w', 0)) > 0
        return len(keys), fs_success, fs_failure, vol_success, vol_failure, warnings


def do_metadata(infinibox, config, count):
    try:
        dryrun = config.general.dryrun
        standalone_volumes_filesystems = config.infinibox.volumes
        if dryrun:
            log.debug('Dryrun, not doing metadata')
            return
        if standalone_volumes_filesystems:
            log.debug('Standalone vol/fs/cg, not doing metadata')
            return
        warnings = has_warnings()

        suppress = config.logging.suppress
        sa_internal = config.general.sa_internal
        force_report = config.general.force_report
        prefix = config.general.snap_basename
        fs_success = count.done_fs_count
        vol_success = count.done_vol_count
        fs_failure = count.todo_fs_count - count.done_fs_count
        vol_failure = count.todo_vol_count - count.done_vol_count

        pool = find_smallest_pool(infinibox)
        current = fill_current_metadata(
            fs_success, fs_failure, vol_success, vol_failure, warnings)
        set_current_metadata(infinibox, pool, prefix, current, suppress)
        history = get_current_metadata(pool, prefix)
        newest, prev = cleanup_metadata(infinibox, pool, history, suppress)
        log.debug("Newest timestamp: {} Previous timestamp: {}", newest, prev)
        p_year, p_mon, p_mday, p_hour, p_min, p_sec, p_wday, p_yday, p_isdst = gmtime(
            prev)
        n_year, n_mon, n_mday, n_hour, n_min, n_sec, n_wday, n_yday, n_isdst = gmtime(
            newest)
        # Send report if day is different from the previous one
        if n_yday != p_yday or p_year != n_year or force_report:
            report_metadata_stats(infinibox, prefix, sa_internal)
    except Exception as e:
        log.uwarn("Updating metadata failed. {}".format(repr(e)))


def fill_current_metadata(fs_success, fs_failure, vol_success, vol_failure, warnings):
    data = dict()
    data['fs'] = fs_success
    data['ff'] = fs_failure
    data['vs'] = vol_success
    data['vf'] = vol_failure
    data['w'] = warnings
    total = dict()
    total['t'] = int(time())
    total['d'] = data
    return total


def clean_old_metadata(pool, days=7):
    log.debug('Cleaning old metadata on pool {}'.format(pool.get_name()))
    all_metadata = list()
    attempts = 0
    while attempts < 5:
        try:
            all_metadata = pool.get_all_metadata()
            break
        except Exception as e:
            attempts += 1
            log.debug(
                "Exception getting metadata. Retrying. {}".format(repr(e)))
            sleep(1)
    filtered_metadata = [a for a in all_metadata if match(
        '^snaprotator_.*\.\d{10}$', a)]
    for metadata_entry in filtered_metadata:
        epoch = int(metadata_entry.split('.')[-1])
        if datetime.now() - timedelta(days=days) > datetime.fromtimestamp(epoch):
            log.debug('Deleting entry older than {} days {}'.format(
                days, metadata_entry))
            try:
                pool.unset_metadata(metadata_entry)
            except Exception as e:
                log.debug('Failed to delete {}. ({})'.format(
                    metadata_entry, repr(e)))


@_retry_on_changed_during_iteration
def find_smallest_pool(infinibox):
    pools = infinibox.pools.to_list()
    log.debug('Pools: {}'.format([a.get_name() for a in pools]))
    sorted_pools = sorted(pools, key=lambda x: x.get_id())
    log.debug('Sorted pools: {}'.format([a.get_name() for a in pools]))
    for pool in sorted_pools:
        clean_old_metadata(pool, days=7)
    for pool in sorted_pools:
        try:
            with infinibox.api.added_headers_context({"X-INFINIDAT-SUPPRESS-EVENTS": "true"}):
                pool.set_metadata('writetest', {})
                pool.unset_metadata('writetest')
                return pool
        except APICommandFailed as e:
            if e.error_code in ('NOT_SUPPORTED_IN_SYSTEM_STATE'):
                log.uabort('Operation not supported in STANDBY mode. Exiting.')
            else:
                log.debug('Error writing metadata. {}'.format(repr(e)))
        except Exception as e:
            log.debug(
                'Failed to set/unset metadata. Pool: {}\n{}'.format(pool.get_name(), repr(e)))
            continue
    log.debug("Failed to find a pool with permissions.")
    return None


def get_current_metadata(pool, policy_name):
    try:
        all_metadata = []
        log.debug('Using pool {}'.format(pool))
        if 'get_name' in dir(pool):
            log.debug('Pool name: {}'.format(pool.get_name()))
            attempts = 0
            while attempts < 5:
                try:
                    all_metadata = pool.get_all_metadata()
                    break
                except Exception as e:
                    attempts += 1
                    log.debug(
                        "Exception getting metadata. Retrying. {}".format(repr(e)))
                    sleep(1)
        log.debug('All metadata: {}'.format(all_metadata))
        filtered_metadata = [
            x for x in all_metadata if 'snaprotator_' + policy_name + '.' in x]
        log.debug('Filtered: {}'.format(filtered_metadata))
        metadata_dict = dict()
        for meta in filtered_metadata:
            metadata_dict[meta] = json.loads(all_metadata[meta])
        log.debug('metadata dict: {}'.format(metadata_dict))
        return metadata_dict
    except Exception as e:
        log.uwarn("No metadata found. Returning default. {}".format(repr(e)))
        return dict()


def set_current_metadata(infinibox, pool, policy_name, metadata, suppress=False):
    if not pool:
        log.debug("No writeable pool found. Skipping setting metadata.")
    else:
        try:
            meta = json.dumps(metadata)
            timestamp = metadata['t']
            with infinibox.api.added_headers_context({"X-INFINIDAT-SUPPRESS-EVENTS": "true"}) if suppress else dummy_context_mgr():
                pool.set_metadata(
                    'snaprotator_' + policy_name + '.' + str(timestamp), meta)
        except Exception as e:
            log.uwarn("Failed to save metadata. {}".format(repr(e)))


def cleanup_metadata(infinibox, pool, metadata, suppress=False):
    if not pool:
        log.debug("No writeable pool found. Skipping cleanup metadata.")
        return 0, 0
    else:
        try:
            keys = sorted([x for x in metadata])
            if 'timestamp' in metadata[keys[-1]]:
                last_timestamp = metadata[keys[-1]]['timestamp']
            else:
                last_timestamp = metadata[keys[-1]]['t']
            prev_timestamp = last_timestamp
            if len(keys) > 1:
                if 'timestamp' in metadata[keys[-2]]:
                    prev_timestamp = metadata[keys[-2]]['timestamp']
                else:
                    prev_timestamp = metadata[keys[-2]]['t']
            log.debug("Latest metadata: {}".format(last_timestamp))
            for entry in keys:
                if 'timestamp' in metadata[entry]:
                    this_timestamp = metadata[entry]['timestamp']
                else:
                    this_timestamp = metadata[entry]['t']
                log.debug("Processing metadata: {}".format(this_timestamp))
                if last_timestamp - this_timestamp > 86400:
                    with infinibox.api.added_headers_context({"X-INFINIDAT-SUPPRESS-EVENTS": "true"}) if suppress else dummy_context_mgr():
                        pool.unset_metadata(entry)
                    log.debug('Successfully deleted metadata {}'.format(entry))
            return last_timestamp, prev_timestamp
        except Exception as e:
            log.uwarn("Failed to cleanup metadata. {}".format(repr(e)))
            return 0, 0


@_retry_on_changed_during_iteration
def infinibox_validate(infinibox, abort=True):
    infinibox_version = None
    if abort:
        log_handler = log.uabort
    else:
        log_handler = log.uwarn
    with infinibox_api_error_wrapper("{}:{}".format(*infinibox.get_api_addresses()[0]), abort):
        infinibox_version = LooseVersion(infinibox.get_version())
        snapshots_count = sum(p.get_field('snapshots_count')
                              for p in infinibox.pools.get_all())
    if not infinibox_version:
        return
    # Validate constraints
    if not MIN_SUPPORTED_VER < infinibox_version < MAX_SUPPORTED_VER:
        log_handler("Infinibox version {} is not supported by this version of snaprotator".format(infinibox_version),
                    exit_code=errors.EXIT_CODE_IBOX_VERSION_UNSUPPORTED)
    if snapshots_count >= MAX_SYSTEM_SNAPSHOTS:
        log_handler("Maximum number of snapshots per system had been reached ({}), rotation will not be performed",
                    MAX_SYSTEM_SNAPSHOTS, exit_code=errors.EXIT_CODE_IBOX_TOO_MANY_SNAPSHOTS_PER_SYSTEM)


@_retry_on_changed_during_iteration
def volume_validate(infinibox, volume_name, is_fs=False, immutability=False):
    log.uinfo("Checking for volume/fs/cg '{}'".format(volume_name))
    log.debug("is_fs={}".format(is_fs))
    infinibox_version = LooseVersion(infinibox.get_version())
    volumes = None
    too_many_snaps = False
    if is_fs is True:
        # Check if filesystem exists with the same name
        # supported at least since 1.7
        if infinibox_version > FS_SUPPORTED_VER:
            if check_count_fs_snaps(infinibox):
                log.uabort("Too many FS snapshots, aborting.",
                           exit_code=errors.EXIT_CODE_IBOX_TOO_MANY_SNAPSHOTS_PER_VOLUME)
            volumes = infinibox.objects.filesystems.find(name=volume_name)
            if not volumes:
                log.uabort("Failed to find filesystem '{}', aborting.".format(volume_name),
                           exit_code=errors.EXIT_CODE_IBOX_VOLUME_NOT_FOUND)
    elif is_fs is False:  # Volumes
        volumes = infinibox.objects.volumes.find(name=volume_name)
        if not volumes:
            # Check if consistency group exists with the same name
            # supported since v. 2.2
            if infinibox_version > CG_SUPPORTED_VER:
                volumes = infinibox.objects.cons_groups.find(
                    name=volume_name)
            if not volumes:
                log.uabort("Failed to find volume/cg '{}', aborting.".format(volume_name),
                           exit_code=errors.EXIT_CODE_IBOX_VOLUME_NOT_FOUND)
    else:
        volumes = infinibox.objects.volumes.find(name=volume_name)
        if not volumes:
            # Check if consistency group exists with the same name
            # supported since v. 2.2
            if infinibox_version > CG_SUPPORTED_VER:
                volumes = infinibox.objects.cons_groups.find(
                    name=volume_name)
            if not volumes:
                if check_count_fs_snaps(infinibox):
                    log.uabort("Too many volume snapshots, aborting.",
                               exit_code=errors.EXIT_CODE_IBOX_TOO_MANY_SNAPSHOTS_PER_VOLUME)
                volumes = infinibox.objects.filesystems.find(name=volume_name)
                if not volumes:
                    log.uabort("Failed to find filesystem/volume/cg '{}', aborting.".format(volume_name),
                               exit_code=errors.EXIT_CODE_IBOX_VOLUME_NOT_FOUND)

    # if iummutability was requested, verifyversion constraints
    if immutability:
        if infinibox_version < IMMUTABLE_SUPPORTED_VER:
            log.uabort("Immutable snaps require a minimum IBOX firmware of "
                       "'{}', aborting.".format(IMMUTABLE_SUPPORTED_VER))

    volume = volumes[0]
    if volume.get_type_name() == 'cons_group':
        snapshots = volume.get_snapgroups()
        if infinibox_version < LooseVersion("6"):
            max_cg_snapshots = MAX_CG_SNAPSHOTS
        else:
            max_cg_snapshots = MAX_CG_SNAPSHOTS_V6

        if len(snapshots) > max_cg_snapshots:
            log.uwarn("Too many snapshots ({}) for CG '{}'.".format(
                len(snapshots), volume_name))
            too_many_snaps = True
    else:
        snapshots = volume.get_snapshots()
        if len(snapshots) > MAX_VOL_SNAPSHOTS:
            log.uwarn("Too many snapshots ({}) for volume '{}'.".format(
                len(snapshots), volume_name))
            too_many_snaps = True

    log.debug("Volume validation - OK")
    return volume, snapshots, too_many_snaps


@_retry_on_changed_during_iteration
def get_all_volumes_by_serial(infinibox):
    return {volume.get_serial(): volume for volume in infinibox.volumes}


@_retry_on_changed_during_iteration
def get_all_volumes(infinibox, allow_replicas=False):
    log.uinfo("Getting all volumes from the system")
    infinibox_version = LooseVersion(infinibox.get_version())
    names_of_volumes_cgs = []
    volumes = infinibox.objects.volumes.find(type='MASTER')
    log.debug("Allow replicas is {}".format(allow_replicas))
    for volume in volumes:
        if infinibox_version > CG_SUPPORTED_VER:
            if volume.get_cons_group() is None:
                if allow_replicas or not volume.is_rmr_target():
                    names_of_volumes_cgs.append(volume)
                else:
                    log.debug(
                        "Volume {} is a replication target, skipping".format(volume.get_name()))
        else:
            if allow_replicas or not volume.is_rmr_target():
                names_of_volumes_cgs.append(volume)
            else:
                log.debug(
                    "Volume {} is a replication target, skipping".format(volume.get_name()))

    if infinibox_version > CG_SUPPORTED_VER:
        cgs = infinibox.objects.cons_groups.find(type='MASTER')
        for cg in cgs:
            replist = cg.get_replicas().to_list()
            target = False
            if len(replist) > 0:
                for replica in replist:
                    if replica.is_target() and not allow_replicas:
                        log.debug(
                            "CG {} is a replication target, skipping".format(cg.get_name()))
                        target = True
            if not target:
                names_of_volumes_cgs.append(cg)

    if not names_of_volumes_cgs:
        log.uinfo("Failed to find volumes/cgs")
    else:
        log.debug("Volume find - OK")
    return names_of_volumes_cgs


@_retry_on_changed_during_iteration
def get_all_snapshots(infinibox, allow_replicas=False, basename=None):
    log.uinfo("Getting all snapshots from the system")
    infinibox_version = LooseVersion(infinibox.get_version())
    snapshots = []
    if infinibox_version > LooseVersion("3.0.0"):
        volumes = [a for a in infinibox.objects.volumes.find(
            type='SNAPSHOT') if '.{}.'.format(basename) in a.get_name()]
    else:
        volumes = [a for a in infinibox.objects.volumes.find(
            type='SNAP') if '.{}.'.format(basename) in a.get_name()]
    log.debug("Allow replicas is {}".format(allow_replicas))
    for volume in volumes:
        if infinibox_version > CG_SUPPORTED_VER:
            if volume.get_cons_group() is None:
                if allow_replicas or not volume.is_rmr_target():
                    snapshots.append(volume)
                else:
                    log.debug(
                        "Volume {} is a replication target, skipping".format(volume.get_name()))
        else:
            if allow_replicas or not volume.is_rmr_target():
                snapshots.append(volume)
            else:
                log.debug(
                    "Volume {} is a replication target, skipping".format(volume.get_name()))

    if infinibox_version > CG_SUPPORTED_VER:
        if infinibox_version > LooseVersion("3.0.0"):
            cgs = [a for a in infinibox.objects.cons_groups.find(
                type='SNAPSHOT') if '.{}.'.format(basename) in a.get_name()]
        else:
            cgs = [a for a in infinibox.objects.cons_groups.find(
                type='SNAP') if '.{}.'.format(basename) in a.get_name()]
        for cg in cgs:
            replist = cg.get_replicas().to_list()
            target = False
            if len(replist) > 0:
                for replica in replist:
                    if replica.is_target() and not allow_replicas:
                        log.debug(
                            "CG {} is a replication target, skipping".format(cg.get_name()))
                        target = True
            if not target:
                snapshots.append(cg)

    if infinibox_version > FS_SUPPORTED_VER:
        if infinibox_version > LooseVersion("3.0.0"):
            filesystems = [a for a in infinibox.objects.filesystems.find(
                type='SNAPSHOT') if '.{}.'.format(basename) in a.get_name()]
        else:
            filesystems = [a for a in infinibox.objects.filesystems.find(
                type='SNAP') if '.{}.'.format(basename) in a.get_name()]
    for snap in filesystems:
        snapshots.append(snap)
    return snapshots


@_retry_on_changed_during_iteration
def check_count_fs_snaps(infinibox):
    log.uinfo("Getting all fs snaps from the system")
    infinibox_version = LooseVersion(infinibox.get_version())
    if infinibox_version > LooseVersion("3.0.0"):
        volumes = infinibox.filesystems.find(type='SNAPSHOT')
    else:
        volumes = infinibox.filesystems.find(type='SNAP')
    log.debug("Total snapshots for FS: {}".format(len(volumes)))
    return len(volumes) >= MAX_FS_SNAPSHOTS


@_retry_on_changed_during_iteration
def get_all_filesystems(infinibox, allow_replicas=False, snap_mapped_fs_snapshots=False):
    log.uinfo("Getting all filesystems from the system")
    infinibox_version = LooseVersion(infinibox.get_version())
    names_of_filesystems = []
    if infinibox_version > FS_SUPPORTED_VER:
        filesystems = infinibox.objects.filesystems.find(type='MASTER')
        if snap_mapped_fs_snapshots:
            filesystems = set(filesystems.to_list(
            ) + infinibox.objects.filesystems.find(mapped=True).to_list())
        for filesystem in filesystems:
            if infinibox_version < LooseVersion("4.0.0") or allow_replicas or not filesystem.is_rmr_target():
                names_of_filesystems.append(filesystem)
            else:
                log.debug(
                    "Volume {} is a replication target, skipping".format(filesystem.get_name()))

    if not names_of_filesystems:
        log.uinfo("Failed to find filesystems")
    else:
        log.debug("Filesystem find - OK")
    return names_of_filesystems


def list_of_volumes_to_snapshot(config, volumes):
    patterns = read_patterns_from_config(config, 'volumes')
    return get_volumes_to_snapshot(volumes, patterns)


def iter_filtered_volumes(volumes, pattern):
    for volume in volumes:
        if fnmatch(volume.lower(), pattern.lower()):
            yield volume


def get_volumes_to_snapshot(volumes, patterns):
    def get_filtered_volumes(pred):
        result = set()
        for pattern in (p for p in patterns if pred(p)):
            if pattern[0] == '!':
                pattern = pattern[1:]
            _res = list(iter_filtered_volumes(volumes, pattern))
            if not _res:
                log.uwarn(
                    "Couldn't find volumes matching pattern {}".format(pattern))
            result.update(_res)
        return result

    includes = get_filtered_volumes(lambda p: p and p[0] != '!')
    excludes = get_filtered_volumes(lambda p: p and p[0] == '!')
    return list(includes - excludes)


def list_of_filesystems_to_snapshot(config, filesystems):
    patterns = read_patterns_from_config(config, 'fs')
    return get_filesystems_to_snapshot(filesystems, patterns)


def iter_filtered_filesystems(filesystems, pattern):
    for filesystem in filesystems:
        if fnmatch(filesystem.lower(), pattern.lower()):
            yield filesystem


def get_filesystems_to_snapshot(filesystems, patterns):
    def get_filtered_filesystems(pred):
        result = set()
        for pattern in (p for p in patterns if pred(p)):
            if pattern[0] == '!':
                pattern = pattern[1:]
            _res = list(iter_filtered_filesystems(filesystems, pattern))
            if not _res:
                log.uwarn(
                    "Couldn't find filesystem matching pattern {}".format(pattern))
            result.update(_res)
        return result

    includes = get_filtered_filesystems(lambda p: p and p[0] != '!')
    excludes = get_filtered_filesystems(lambda p: p and p[0] == '!')
    return list(includes - excludes)


@timed
def do_infinibox_snapshot(infinibox, config, volume_name, is_fs=False):
    remote_snap = None
    if not is_fs and config.general.snap_both_aa_peers:
        volumes = infinibox.objects.volumes.find(
            name=volume_name, type='MASTER').to_list()
        cgs = infinibox.objects.cons_groups.find(
            name=volume_name, type='MASTER').to_list()
        all_objects = volumes + cgs
        if all_objects:
            volume = all_objects[0]
            volume_type = volume.get_type_name()
            log.uinfo('Checking {} {}'.format(volume_type, volume.get_name()))
            is_active_active = False
            rep_type = volume.get_replication_types()[0]
            if volume.is_replicated():
                if rep_type in ['ACTIVE_ACTIVE', 'SYNC_SOURCE']:
                    is_active_active = True

            if is_active_active:
                try:
                    log.uinfo("{} {} is ACTIVE_ACTIVE".format(
                        volume_type.capitalize(), volume.get_name()))
                    if volume_type == 'cons_group':
                        base = [a for a in volume.get_replicas(
                        ).to_list() if a.is_type_active_active()][0]
                    else:
                        base = volume.get_replica()

                    link_id = base.get_link().get_id()
                    res = infinibox.api.get(
                        "/api/rest/remote/{}/api/rest/config".format(link_id))
                    ip = res.get_json()['result']['ip_config']['ip']
                    remote_username, remote_password = infinibox.get_api_auth()
                    remote_config = {
                        'infinibox': {
                            'host': ip,
                            'username': remote_username,
                            'password': remote_password
                        }
                    }
                    remote_infinibox = connect_to_infinibox(
                        remote_config, abort=False)
                    infinibox_validate(remote_infinibox, abort=False)
                    infinibox.register_related_system(remote_infinibox)
                    remote_volume_name = base.get_remote_entity().get_name()
                    remote_snap, remote_metadata = _do_infinibox_snapshot(
                        remote_infinibox, config, remote_volume_name, is_fs)
                    if remote_snap:
                        log.uinfo('Successfully created remote snapshot on {}/{}'.format(
                            remote_infinibox.get_name(), remote_volume_name))
                except Exception as e:
                    log.uwarn(repr(e))
                    log.uwarn("Couldn't create remote snapshot. Continuing.")

    local_snap, metadata = _do_infinibox_snapshot(
        infinibox, config, volume_name, is_fs)
    if local_snap:
        log.uinfo('Successfully created local snapshot on {}/{}'.format(
            infinibox.get_name(), volume_name))

    return local_snap, metadata, remote_snap


def _do_infinibox_snapshot(infinibox, config, volume_name, is_fs=False):
    basename = config.general.snap_basename
    delete = config.general.delete
    lock_days = config.general.snap_lock_days or 0

    volume, snapshots, too_many_snaps = volume_validate(infinibox, volume_name,
                                                        is_fs, True if lock_days else False)
    if not delete:
        if not too_many_snaps:
            snapshot_obj = create_latest_snapshot(
                infinibox, config, volume)
            # We created a snapshot, and it's not in the list we pass
            snapshots = [s for s in snapshots] + [snapshot_obj]
            config.general.actual_retain = int(config.general.snaps_to_keep)
            deleted_snaps_metadata = trim_number_of_snapshots(
                infinibox, config, snapshots, volume)
        else:
            log.uwarn('Too many snaps, deleting snap before creating')
            relevant_snaps = filter_irrelevant_snapshots(
                snapshots, volume.get_name(), basename)
            log.uwarn('Will keep {} (required = {})'.format(
                int(len(relevant_snaps)), config.general.snaps_to_keep))
            config.general.actual_retain = int(len(relevant_snaps) - 1)
            deleted_snaps_metadata = trim_number_of_snapshots(
                infinibox, config, snapshots, volume)
            snapshot_obj = create_latest_snapshot(
                infinibox, config, volume)
    else:
        snapshot_obj = None
        config.general.snaps_to_keep = 0
        deleted_snaps_metadata = trim_number_of_snapshots(
            infinibox, config, snapshots, volume)
    return snapshot_obj, deleted_snaps_metadata


@timed
def do_purge(infinibox, config, volume_name, is_fs=False):
    volume, snapshots, too_many_snaps = volume_validate(
        infinibox, volume_name, is_fs)
    config.general.snaps_to_keep = 0
    deleted_snaps_metadata = trim_number_of_snapshots(
        infinibox, config, snapshots, volume)
    # fixme add vmware deletion
    return deleted_snaps_metadata


def purge_volumes_filesystems(infinibox, config, volumes, filesystems):
    for volume in volumes:
        do_purge(infinibox, config,
                 volume, is_fs=False)
    for filesystem in filesystems:
        do_purge(
            infinibox, config, filesystem, is_fs=True)


def find_wildcard_volumes_fs(infinibox, config):
    allow_replicas = config.general.get('allow_replicas', False)
    snap_mapped_fs_snapshots = config.get(
        'general').get('snap_mapped_fs_snapshots', False)

    all_volume_objects = get_all_volumes(infinibox, allow_replicas)
    all_volumes = [vol.get_name() for vol in all_volume_objects]
    standalone_volumes = list_of_volumes_to_snapshot(config, all_volumes)
    log.uinfo(
        'The following volumes will be snapshotted: {}', [str(a) for a in standalone_volumes])
    all_filesystem_objects = get_all_filesystems(
        infinibox, allow_replicas, snap_mapped_fs_snapshots)
    all_filesystems = [fs.get_name()
                       for fs in all_filesystem_objects]
    standalone_filesystems = list_of_filesystems_to_snapshot(
        config, all_filesystems)
    log.uinfo(
        'The following filesystems will be snapshotted: {}', [str(a) for a in standalone_filesystems])
    return standalone_volumes, standalone_filesystems


def find_specific_volumes_fs(infinibox, config):
    config_data = config.infinibox.volumes
    lock_days = config.general.snap_lock_days or 0

    standalone_volumes, standalone_filesystems = list(), list()

    if config_data:
        standalone_volumes_filesystems = config_data.split(',')
        log.uinfo('The following volumes will be snapshotted: {}', [
            str(a) for a in standalone_volumes_filesystems])
        for volume_name in standalone_volumes_filesystems:
            volume, snapshots, too_many_snaps = volume_validate(infinibox, volume_name,
                                                                'unknown', True if lock_days else False)
            volume_type = volume.get_type_name()
            if volume_type == 'filesystem':
                standalone_filesystems.append(volume_name)
            else:
                standalone_volumes.append(volume_name)

    return standalone_volumes, standalone_filesystems


@timed
def do_infinibox_volumes(infinibox, config, count, volumes):
    count.done_vol_count = count.todo_vol_count
    for volume in volumes:
        try:
            snapshot_name, metadata_list, remote_snapshot = do_infinibox_snapshot(
                infinibox, config, volume, is_fs=False)
            if not snapshot_name:
                count.done_vol_count -= 1
        except SystemExit as e:
            count.done_vol_count -= 1
            if e.code == errors.EXIT_CODE_IBOX_VOLUME_NOT_FOUND:
                log.uwarn(
                    'Infinibox volume {} not found, skipping it.', volume)
                continue
            else:
                raise


@timed
def do_infinibox_filesystems(infinibox, config, count, filesystems):
    count.done_fs_count = count.todo_fs_count
    for filesystem in filesystems:
        try:
            snapshot_name, metadata_list, remote_snapshot = do_infinibox_snapshot(
                infinibox, config, filesystem, is_fs=True)
            if not snapshot_name:
                count.done_fs_count -= 1
        except SystemExit as e:
            count.done_fs_count -= 1
            if e.code == errors.EXIT_CODE_IBOX_VOLUME_NOT_FOUND:
                log.uwarn(
                    'Infinibox filesystem {} not found, skipping it.', filesystem)
                continue
            else:
                raise


def send_custom_event(infinibox, level, visibility, description_template, data=None):
    if infinibox is not None:
        infinibox.events.create_custom_event(level=level, description=description_template,
                                             visibility=visibility, data=data)
    else:
        log.debug("CUSTOM_EVENT LEVEL={level} VISIBILITY={visibility} description_template={description_template!r} data={data!r}".format(
            level=level, visibility=visibility, description_template=description_template, data=data))


def event_locked(infinibox, sa_internal=False):
    template = "Another snaprotator process is running"
    subsystem_type = "external"
    subsystem_code = "CUSTOMER_SNAPROTATOR"
    if sa_internal:
        subsystem_type = "internal"
        subsystem_code = "SA_SNAPROTATOR"
    data = [
        dict(name="snaprotator_version",
             type="string", value=snaprotator_version),
        dict(name="subsystem_type", type="string", value=subsystem_type),
        dict(name="subsystem_code", type="string", value=subsystem_code),
    ]
    send_custom_event(infinibox, level="WARNING",
                      visibility="CUSTOMER", description_template=template, data=data)


def event_24_h(infinibox, runs, policy, fs_success, fs_failure, vol_success, vol_failure, warnings, sa_internal=False):
    level = "INFO"
    # if vol_failure > 0 or fs_failure > 0:
    #     level = "ERROR"
    # elif warnings:
    #     level = "WARNING"
    subsystem_type = "external"
    subsystem_code = "CUSTOMER_SNAPROTATOR"
    if sa_internal:
        subsystem_type = "internal"
        subsystem_code = "SA_SNAPROTATOR"

    template = 'Snaprotator 24h status for policy "{policy}": {level}. Total {runs} runs. FS success: {fs_success} FS failure: {fs_failure} VOL/CG Success: {vol_success} VOL/CG failure: {vol_failure}. Runs with warnings: {warnings}'
    data = [
        dict(name="policy", type="string", value=policy),
        dict(name="runs", type="string", value=runs),
        dict(name="level", type="string", value=level),
        dict(name="warnings", type="string", value=warnings),
        dict(name="fs_success", type="string", value=fs_success),
        dict(name="fs_failure", type="string", value=fs_failure),
        dict(name="vol_success", type="string", value=vol_success),
        dict(name="vol_failure", type="string", value=vol_failure),
        dict(name="subsystem_type", type="string", value=subsystem_type),
        dict(name="subsystem_code", type="string", value=subsystem_code),
        dict(name="snaprotator_version",
             type="string", value=snaprotator_version),
    ]
    send_custom_event(infinibox, level=level,
                      visibility="CUSTOMER", description_template=template, data=data)


def event_unknown_error(infinibox, sa_internal=False):
    template = "Snaprotator finished with status: ERROR"
    subsystem_type = "external"
    subsystem_code = "CUSTOMER_SNAPROTATOR"
    if sa_internal:
        subsystem_type = "internal"
        subsystem_code = "SA_SNAPROTATOR"
    data = [
        dict(name="snaprotator_version",
             type="string", value=snaprotator_version),
        dict(name="subsystem_type", type="string", value=subsystem_type),
        dict(name="subsystem_code", type="string", value=subsystem_code),
    ]
    send_custom_event(infinibox, level="ERROR",
                      visibility="CUSTOMER", description_template=template, data=data)
