import os
import sys
import logbook
from logbook.more import ColorizedStderrHandler
from .errors import EXIT_CODE_GENERAL_ERROR

_has_warnings = 0


def has_warnings():
    return _has_warnings


class UserLogger(logbook.Logger):

    """docstring for UserLogger"""

    def uinfo(self, *args, **kwargs):
        self.info(*args, user=True, **kwargs)

    def uwarn(self, *args, **kwargs):
        global _has_warnings
        _has_warnings += 1
        self.warn(*args, user=True, **kwargs)

    def uabort(self, *args, **kwargs):
        self.error(*args, user=True, **kwargs)
        exit_code = kwargs.get('exit_code', EXIT_CODE_GENERAL_ERROR)
        raise SystemExit(exit_code)

    def uabort_with_exc_info(self, *args, **kwargs):
        self.error(*args, user=True, exc_info=True, **kwargs)
        exit_code = kwargs.get('exit_code', EXIT_CODE_GENERAL_ERROR)
        raise SystemExit(exit_code)


log = UserLogger(__name__)


def setup_file_handler(config):
    try:
        file_path = config['logging']['logpath']
        if not os.path.exists(os.path.dirname(file_path)):
            os.makedirs(os.path.dirname(file_path))

        if config['logging'].get('rotation', True):
            logs_to_keep = config.get('logging', {}).get('logs_to_keep', 14)
            date_format = '%Y-%m-%d'
            file_handler = logbook.TimedRotatingFileHandler(
                file_path, mode='a', date_format=date_format, backup_count=logs_to_keep, level=logbook.DEBUG, bubble=True)
            cleanup = file_handler.files_to_delete()
            for date_file in cleanup:
                date, filename = date_file
                os.remove(filename)
        else:
            file_handler = logbook.MonitoringFileHandler(
                file_path, level=logbook.DEBUG, bubble=True)

    except Exception:
        msg = "caught exception when creating logfile {}".format(file_path)
        err = sys.exc_info()[1]
        sys.stderr.write('{msg} ({err})\n'.format(msg=msg, err=err))
        log.uabort_with_exc_info(msg)

    return file_handler


def setup_email(config):
    class MyMailHandler(logbook.MailHandler):

        def __init__(self, *args, **kwargs):
            self.policy = None
            # Output email in report radar format
            self.radar = False
            self.infinibox = None
            self.level = None
            # Send email only on Warning and Failure
            self.email_skip_success = False
            logbook.MailHandler.__init__(self, *args, **kwargs)

        def emit_batch(self, records, _):
            try:
                if not records:
                    return
                max_level = max(r.level for r in records)
                if max_level >= logbook.ERROR:
                    self.level = 'Error'
                    self.subject = "{}ERROR: snaprotator: Failed ({})".format(
                        f"Job: {emailsubject} - " if emailsubject else "", self.policy)
                elif max_level >= logbook.WARNING:
                    self.level = 'Warning'
                    self.subject = "{}WARNING: snaprotator: Some operations failed ({})".format(
                        f"Job: {emailsubject} - " if emailsubject else "", self.policy)
                else:
                    if self.email_skip_success:
                        return
                    self.level = 'Success'
                    self.subject = "{}SUCCESS: snaprotator: Finished successfully ({})".format(
                        f"Job: {emailsubject} - " if emailsubject else "", self.policy)
            
                return super(MyMailHandler, self).emit_batch(records, 'escalation')
            except Exception as e:
                sys.stderr.write(
                    'Error sending mail. Please check your configuration file. ({})\n'.format(repr(e)))

    mail_format_string = '''\
    Subject: {handler.subject}

    Message type:       {record.level_name}
    Location:           {record.filename}:{record.lineno}
    Module:             {record.module}
    Function:           {record.func_name}
    Time:               {record.time:%Y-%m-%d %H:%M:%S}

    Message:

    {record.message}
    '''
    backup_radar_email_format_string = '''\
    Subject: Backup Report {handler.subject}

    Device: {handler.infinibox}
    Status: {handler.level}
    Job: {handler.policy}

    Message type:       {record.level_name}
    Location:           {record.filename}:{record.lineno}
    Module:             {record.module}
    Function:           {record.func_name}
    Time:               {record.time:%Y-%m-%d %H:%M:%S}

    Message:

    {record.message}
    '''

    mail_related_format_string = '{record.time:%Y-%m-%d %H:%M:%S}:{record.level_name}:{record.message}'

    sender = config.logging.emailsender
    rcpt = config.logging.emailrcpt
    emailserver = config.logging.emailserver

    if 'emailport' in config.logging:
        emailport = int(config.logging.emailport)
    else:
        emailport = 25
    
    emailsubject = config.logging.emailsubject

    kargs = {}
    kargs['from_addr'] = sender
    kargs['recipients'] = rcpt
    kargs['subject'] = emailsubject
    kargs['server_addr'] = (emailserver, emailport)
    kargs['level'] = 'INFO'

    kargs['related_format_string'] = mail_related_format_string
    kargs['bubble'] = True

    if config.logging.emailuser:
        username = config.logging.emailuser
        password = config.logging.emailpassword
        kargs['credentials'] = (username, password)
        kargs['secure'] = ()
    backup_radar_email_format = config.logging.backup_radar_email_format
    if backup_radar_email_format:
        kargs['format_string'] = backup_radar_email_format_string
    else:
        kargs['format_string'] = mail_format_string
    email_handler = MyMailHandler(**kargs)
    email_handler.policy = config.general.snap_basename
    email_handler.backup_radar_email_format = backup_radar_email_format
    email_handler.infinibox = config.infinibox.host
    email_handler.email_skip_success = config.logging.email_skip_success 
    return email_handler


def setup_logging(config):
    logbook.set_datetime_format('local')
    loggers = []

    class NoExceptionStringFormatter(logbook.StringFormatter):

        def format_exception(self, _):
            return None

    console_level = logbook.INFO

    def console_filter(r, _): return r.kwargs.get('user', False)
    console_log_format = (
        '{record.time:%H:%M}\t'
        '{record.level_name}\t'
        '{record.message}'
    )
    if 'logging' in config:
        # Set up logging levels
        if config.get('logging', {}).get('debug'):
            console_level = logbook.DEBUG

            def console_filter(r, _): return True
            console_log_format = logbook.StringFormatterHandlerMixin.default_format_string
        elif config.logging.quiet:
            console_level = logbook.ERROR
            console_log_format = '{record.message}'

        # Add logfile to handlers if requested
        if config.logging.logpath:
            file_handler = setup_file_handler(config)
            loggers.append(file_handler)

        try:
            if config.logging.syslog:
                syslog_handler = logbook.SyslogHandler(
                    application_name=log.name,
                    level=logbook.DEBUG,
                    bubble=True
                )
                loggers.append(syslog_handler)
        except Exception as e:
            sys.stderr.write('Could not add Syslog Handler')

    stderr_handler = ColorizedStderrHandler(
        level=console_level,
        filter=console_filter,
        bubble=True
    )
    stderr_handler.formatter = NoExceptionStringFormatter(console_log_format)
    loggers.append(stderr_handler)

    return logbook.NestedSetup(loggers)


def setup_mail_logging(config, dryrun):
    if not dryrun and config.logging.email:
        email_handler = setup_email(config)
        return logbook.GroupHandler(email_handler)
    else:
        return logbook.NullHandler()
