import ssl
import json
import requests
import sys
from . import vmware
from .logsetup import UserLogger
from distutils.version import LooseVersion
from urllib.parse import urlparse

MIN_POWERTOOLS_VERSION = LooseVersion("5.2.1")
MIN_POWERTOOLS_SUPPRESS_METADATA_VERSION = LooseVersion("6.1")


requests.packages.urllib3.disable_warnings()

try:
    _create_unverified_https_context = ssl._create_unverified_context
except AttributeError:
    # Legacy Python that doesn't verify HTTPS certificates by default
    pass
else:
    # Handle target environment that doesn't support HTTPS verification
    ssl._create_default_https_context = _create_unverified_https_context

log = UserLogger(__name__)


def http(url, data):
    log.debug('http: url  {}'.format(url))
    log.debug('http: data {}'.format(data))
    try:
        r = requests.post(url, data=data, allow_redirects=True, timeout=60)
        return True, r.content
    except Exception as e:
        err = sys.exc_info()[0]
        log.uwarn('Failed HTTP request. url: {}, data: {}, err: {}',
                  url, data, repr(e))
        return False, err


def delete_snapshot(powertools_info, vm_refs, snap_uuid, vm_uuid):
    host_power_tools_vm_ip, session_key, powertools_version = powertools_info
    log.debug("Snap UUID: {}", snap_uuid)
    log.debug("vm_refs: {}", vm_refs)

    url = 'http://%s/plugin/views/virtualmachine/operations/remove_snapshot?moref=Folder:group-d1&uuid=%s&locale=en_US&vimSessionKey=%s' %\
        (host_power_tools_vm_ip, vm_uuid, session_key)
    data = {"snapshot_id": snap_uuid, "got_credentials": True}
    success, resp = http(url, data)
    log.debug('Got response: {}', resp)
    if not success:  # http request failed, we don't need to retry
        return False, resp
    json_response = {}
    try:
        json_response = json.loads(resp)
    except Exception as e:
        log.debug(
            'Failed to delete snapshot from Host Power Tools for VMWare {}'.format(repr(e)))
        log.debug('{}'.format(resp))
    if 'success' in json_response and json_response['success']:
        return True, json_response

    log.uwarn(
        'Can\'t delete snapshot from Host Power Tools for VMWare. Please delete the snapshot manually.')
    return False, json_response


def delete_snapshot_with_metadata(powertools_info, metadata):
    original_datastore = None
    snapshot_name = None
    if 'original_datastore' in metadata and 'name' in metadata:
        original_datastore = metadata['original_datastore']
        snapshot_name = metadata['name']

    if 'vm_refs' in metadata and 'first_vm_uuid' in metadata and 'first_vm_snapshot_id' in metadata:
        vm_refs = metadata['vm_refs']
        snap_uuid = metadata['first_vm_snapshot_id']
        vm_uuid = metadata['first_vm_uuid']
        success, myerror = delete_snapshot(powertools_info,
                                           vm_refs, snap_uuid, vm_uuid)
        if success:
            log.uinfo('Successfuly removed snapshot ({}) from VMware Power Tools (ds: {})',
                      snapshot_name, original_datastore)
            return True, "Success"
        else:
            log.uwarn('Failed to remove snapshot ({}) from VMware Power Tools (ds: {})({})',
                      snapshot_name, original_datastore, myerror)
            return False, myerror

    return True, "Can't find supported metadata.Either no vms found on datastore or check VMWare Host Power Tools version."


def get_all_vms_string(vms):
    vmstring = ''
    for vm in vms:
        if vm.summary.config.uuid is None:
            continue
        vmstring = vmstring + vmware.get_vm_string_name(vm) + ";"
    return vmstring


def get_all_vms_names(vms):
    return [vm.name for vm in vms]


def register_snapshot(powertools_info, vms_infinisnapid_datastore, guid, suppress_metadata):
    def get_serial_snapid(snapshot):
        serial = snapshot.system.get_serial()
        snapid = snapshot.get_id()
        return {"system_serial": serial, "snapshot_id": snapid}

    host_power_tools_vm_ip, session_key, powertools_version = powertools_info
    vms, infinisnap, datastore, remote_snap = vms_infinisnapid_datastore
    vm_data, vmnames = vms
    vmstring = ";".join([a.get("moref") for a in vm_data])
    infinisnapid = infinisnap.get_id()
    datastorename = vmware.get_datastore_string_name(datastore)
    if vmstring == '':
        return True, "No vms found on datastore"
    url = 'http://%s/plugin/views/datastore/register_snapshot?moref=%s&locale=en_US&vimSessionKey=%s' %\
        (host_power_tools_vm_ip, datastorename, session_key)
    if powertools_version < MIN_POWERTOOLS_VERSION:
        log.debug("Powertools version is less than {}".format(
            MIN_POWERTOOLS_VERSION))

        data = {"snapshot_id": infinisnapid, "vmrefs": vmstring}
    else:
        log.debug("Powertools version is newer than {}".format(
            MIN_POWERTOOLS_VERSION))

        infinibox_data = [get_serial_snapid(infinisnap)]
        if remote_snap:
            infinibox_data.append(get_serial_snapid(remote_snap))
        data = {
            "guid": guid,
            "vms": vm_data,
            "infinibox_snapshots": infinibox_data,
        }
    if powertools_version >= MIN_POWERTOOLS_SUPPRESS_METADATA_VERSION and suppress_metadata:
        log.debug("Adding suppress metadata option")
        url += "&suppress_events=true"

    success, resp = http(url, json.dumps(data))
    if not success:
        return False, "HTTP request failed"

    try:
        json_response = json.loads(resp)
    except Exception as e:
        log.uwarn(
            'Failed to register snapshot with Host Power Tools for VMware {}', repr(e))
        return False, resp
    if 'success' not in json_response or not json_response['success']:
        if 'error_message' in json_response:
            err = json_response['error_message']
        else:
            err = json_response
        log.uwarn('Failed to register snapshot with Host Power Tools for VMware {} ({})',
                  datastore.name, str(err))
        return False, err
    else:
        log.uinfo(
            'Successfully registered snapshot with Host Power Tools for VMware for {}', datastore.name)
        log.uinfo('Registered VMS: {}', vmnames)
        return True, 'Success'


def get_powertools_info(client):
    url = vmware.find_vmware_power_tools_url(client)
    urlparts = urlparse(url)
    host_power_tools_vm_ip = urlparts.netloc
    session_key = client.session_manager.currentSession.key
    from xml.etree import cElementTree as ET
    try:
        if url and session_key:
            data = {}
            success, resp = http(url, data)
            if success:
                try:
                    root = ET.fromstring(resp)
                    powertools_version = root.attrib.get('version')
                    log.uinfo("Host Power Tools Version: {}".format(
                        powertools_version))
                except Exception as e:
                    pass
                if not powertools_version:
                    return None
                return (host_power_tools_vm_ip, session_key, LooseVersion(powertools_version))
        else:
            return None
    except Exception as e:
        log.uwarn('{}'.format(e))
        return None


def register_snap_unregister_snaps(powertools_info, vms_infinisnapid_datastore, deleted_snaps_metadata_list, guid, suppress_metadata):
    if vms_infinisnapid_datastore:
        success, err = register_snapshot(
            powertools_info, vms_infinisnapid_datastore, guid, suppress_metadata)
        if success:
            log.debug(err)
        else:
            log.uwarn('Failed to register. ({})', str(err))

    for metadata in deleted_snaps_metadata_list:
        success, err = delete_snapshot_with_metadata(powertools_info, metadata)
        if success:
            log.debug(err)
        else:
            log.uwarn('Failed to unregister. ({})', str(err))
