"""

Usage:
   snaprotator [options]
   snaprotator -h | --help
   snaprotator --version


Options:
   -c configfile --configfile=<configfile>        Config file location
   -H hostname --hostname=<hostname>              DNS name of the infinibox
   -u user --user=<username>                      Administrator's username
   -p password --password=<password>              Administrator's password
   -v volume1[,volume2...] --volumes=<volume1[,volume2...>
                                                  Volume(s) to rotate the snaphsots on
   -b basename --basename=<basename>              Basename (daily,weekly, etc...)
   -r retain --retain=<retain>                    Number of snapshots to retain
   --lockdays=<snap_lock_days>                    Number of days to lock snapshot
   --maxlockdays=<max_snap_lock_days>             Max number of days a snap can be locked
   -l logfile --logfile=<logfile>                 Use log file instead of stderr
   -h, --help                                     Show this screen
   -D, --debug                                    Debug
   -q, --quiet                                    Suppress all informational messages
   -d, --dryrun                                   Show what will be done
   -P, --purge                                    Purge all snapshots with basename <basename>
   -s, --syslog                                   Log to syslog daemon
   -n, --norotation                               Do NOT rotate logfile, keep writing indefinitely
   -V, --version                                  Show snaprotator version
   -Z, --force_report                             Force heartbeat (last 24 hours report)
   -R, --delete                                   Delete all snaps with given basename using wildcard
   -L, --list                                     Show all snapshots belonging to policy
   -J, --listjson                                 Show all snapshots belonging to policy in JSON format
   -e password --encrypt=<password>               Encrypt cleartext password for config file use
"""
import signal
import sys
import logbook
from snaprotator.logsetup import UserLogger, setup_logging, setup_mail_logging, has_warnings

from docopt import docopt
from snaprotator.__version__ import __version__ as snaprotator_version

from snaprotator.vmware import do_infinibox_vmware_snapshots
from snaprotator import config as conf
from snaprotator.infinibox import (
    do_metadata,
    connect_to_infinibox,
    infinibox_validate,
    event_unknown_error,
    event_locked,
    get_all_snapshots,
    find_specific_volumes_fs,
    find_wildcard_volumes_fs,
    purge_volumes_filesystems,
    do_infinibox_volumes,
    do_infinibox_filesystems
)

from snaprotator import errors
from collections import OrderedDict
import json
from shutil import copyfile
import os
from snaprotator.utils import (locking_context,
                               signal_handler,
                               to_defaultdict,
                               mask,
                               dummy_context_mgr)

log = UserLogger(__name__)

MAX_ATTEMPTS = 3
RETRY_IN = 5
MAX_RETAIN = 1000


class Count(object):
    """docstring for Count"""

    def __init__(self):
        super(Count, self).__init__()
        self.todo_vol_count = 0
        self.todo_fs_count = 0
        self.done_vol_count = 0
        self.done_fs_count = 0


def copy_example_config():
    cwd = os.path.dirname(__file__)
    example_config = '{}/snaprotator.conf'.format(cwd)
    destination = '/etc/snaprotator.conf.example'
    try:
        copyfile(example_config, destination)
    except Exception:
        pass


def list_snapshots(infinibox, config):
    basename = config.general.get('snap_basename')
    allow_replicas = config.general.get('allow_replicas', False)
    listplain = config.general.get('list', False)

    snapshots = get_all_snapshots(infinibox, allow_replicas, basename)
    mydic = OrderedDict()

    for snap in snapshots:
        entry = dict()
        name = snap.get_name()
        entry['typename'] = typename = snap.get_type_name()
        entry['parent'] = snap.get_parent().get_name()
        entry['size'] = str(
            snap.get_size()) if typename != 'cons_group' else '<N/A>'
        entry['used'] = str(
            snap.get_used_size()) if typename != 'cons_group' else '<N/A>'
        mydic[name] = entry
    # print mydic, using print to output to stdin
    if listplain:
        print(('{:30} {:10} {:10} {:10} {:28}'.format(
            'Name', 'Type', 'Size', 'Used', 'Parent')))
        for key, val in mydic.items():
            print(('{:30} {:10} {:10} {:10} {:28}'.format(
                key, val['typename'], val['size'], val['used'], val['parent'])))
    else:
        print((json.dumps(mydic)))


def initialize_todo(config, filesystems, volumes):
    sa_internal = config.general.get('sa_internal', False)
    count = Count()

    if not filesystems and not volumes:
        if sa_internal:
            log.uwarn(
                "No filesystems/volumes/cgs found.")
            raise SystemExit(
                errors.EXIT_CODE_NOTHING_TODO)
        else:
            log.uabort_with_exc_info(
                "No filesystems/volumes/cgs found. Aborting.")
    count.todo_fs_count = len(filesystems)
    count.todo_vol_count = len(volumes)
    log.debug(
        "Total of {} fs, {} volumes to snap".format(count.todo_fs_count, count.todo_vol_count))
    return count


def report_warnings(config, count):
    if has_warnings():
        log.uinfo(
            'Found {} warnings'.format(has_warnings()))
        log.uinfo('Done (with warnings). FS: {} out of {}, Vol: {} out of {}'.format(
            count.done_fs_count, count.todo_fs_count, count.done_vol_count, count.todo_vol_count))
        raise SystemExit(
            errors.EXIT_CODE_WARNINGS_FOUND)
    else:
        log.uinfo('Finished without warnings or errors. FS: {} out of {}, Vol: {} out of {}'.format(
            count.done_fs_count, count.todo_fs_count, count.done_vol_count, count.todo_vol_count))
        raise SystemExit(
            errors.EXIT_CODE_SUCCESS)


def main():
    signal.signal(signal.SIGINT, signal_handler)
    arguments = docopt(__doc__, version=snaprotator_version)
    if arguments.get('--encrypt'):
        sys.stdout.write('{}\n'.format(mask(arguments.get('--encrypt'))))
        sys.exit(0)
    config = conf.load_config(arguments) if arguments['--configfile'] else {}
    conf.update_arguments_to_config(arguments, config)
    if not config:
        docopt(__doc__, '-h')

    validate, out = conf.validate_config(config)
    if not validate:
        sys.stderr.write(
            'Configuration file or command line parameters error: {}\n'.format(out))
        sys.exit(errors.EXIT_CODE_CONFIG_ERROR)
    config = to_defaultdict(out)

    purge = config.get('infinibox').get('purge', False)
    dryrun = config.general.get('dryrun', False)
    is_vmware = 'vcenter' in config
    single_instance = config.general.get('single_instance', False)
    sa_internal = config.general.get('sa_internal', False)
    delete = config.general.get('delete', False)
    listplain = config.general.get('list', False)
    listjson = config.general.get('listjson', False)
    if listjson or listplain:
        config.logging.quiet = True

    log_handler = setup_logging(config)

    mail_handler = setup_mail_logging(config, dryrun)
    infinibox = None
    with logbook.NullHandler().applicationbound(), mail_handler.applicationbound():
        with log_handler.applicationbound():
            try:
                log.uinfo('SnapRotator v.{}'.format(snaprotator_version))
                if dryrun:
                    log.uinfo('DRYRUN')
                if not dryrun and delete:
                    log.uinfo(
                        'Deleting snapshots with basename {}'.format(config.general.snap_basename))
                infinibox = connect_to_infinibox(config)
                infinibox_validate(infinibox)

                try:
                    with locking_context("/tmp/snaprotator.pid") if single_instance else dummy_context_mgr():
                        if listjson or listplain:
                            list_snapshots(infinibox, config)
                        else:
                            if is_vmware:
                                count = Count()
                                count.todo_vol_count, count.done_vol_count = do_infinibox_vmware_snapshots(
                                    infinibox, config)
                            else:
                                if config.infinibox.volumes:
                                    volumes, filesystems = find_specific_volumes_fs(
                                        infinibox, config)
                                else:
                                    volumes, filesystems = find_wildcard_volumes_fs(
                                        infinibox, config)
                                count = initialize_todo(config,
                                                        filesystems,
                                                        volumes)

                                if purge:
                                    purge_volumes_filesystems(
                                        infinibox, config, volumes, filesystems)
                                else:  # infinibox only
                                    do_infinibox_volumes(
                                        infinibox, config, count, volumes)
                                    do_infinibox_filesystems(
                                        infinibox, config, count, filesystems)
                            report_warnings(config, count)

                except SystemExit:
                    raise
                except Exception as e:
                    log.uwarn("{}".format(repr(e)))
                    raise SystemExit(errors.EXIT_CODE_GENERAL_ERROR)

            except SystemExit as e:
                if not dryrun:
                    if e.code in [errors.EXIT_CODE_WARNINGS_FOUND,
                                  errors.EXIT_CODE_NOTHING_TODO,
                                  errors.EXIT_CODE_SUCCESS]:
                        do_metadata(infinibox, config, count)
                    if (e.code == errors.EXIT_CODE_SUCCESS):
                        log.uinfo("Done.")
                    elif (e.code == errors.EXIT_CODE_WARNINGS_FOUND):
                        log.uwarn("Done with warnings.")
                    elif (e.code == errors.EXIT_CODE_NOTHING_TODO):
                        log.debug(
                            "No fs/cg/vol found, but its ok - sa_internal")
                        log.uinfo("Done. Nothing to do.")
                    elif (e.code == errors.EXIT_CODE_ALREADY_RUNNING):
                        event_locked(infinibox, sa_internal)
                        log.uwarn("Done with error(s).")
                    else:
                        event_unknown_error(infinibox, sa_internal)
                        log.uwarn("Done with error(s).")
                sys.exit(e.code)
            except Exception as e:
                log.uwarn("{}".format(repr(e)))
                log.debug("Unknown Error")
                if not dryrun:
                    event_unknown_error(infinibox, sa_internal)
                log.uabort_with_exc_info(
                    "Internal error occurred, please contact Infinidat support.")
            sys.exit(0)
