from contextlib import contextmanager, redirect_stdout, redirect_stderr
from infinisim import Infinibox as Simulator
from infinisdk import InfiniBox
from infinisdk_internal import enable
import ecosystem
from capacity import GiB, TiB
from unittest import TestCase
from unittest.mock import patch
import snaprotator.snaprotator
import snaprotator.config
import snaprotator.infinibox
import snaprotator.vmware
from infi.pyvmomi_wrapper import Client
from mock import Mock, MagicMock
import glob
import sys
import io
import os
import re
import logbook
from time import sleep, time
import json


class TestRotation(TestCase):
    VOL_PREFIX = 'testvol'
    FS_PREFIX = 'my.fs'
    CREATION_DATES = ['2014.10.21_160801', '2014.10.22_010801',
                      '2014.10.22_050809', '2014.10.22_130104',
                      '2014.10.22_150809', '2014.10.22_152931',
                      '2014.10.22_160801', '2014.10.22_180801']
    BASENAMES = ['daily', VOL_PREFIX]
    MAXVOLS = 5
    NAME_TEMPLATE = "{volname}.{basename}.{date}"
    TEST_NAME_TEMPLATES = [NAME_TEMPLATE,
                           "{volname}.{basename}.{date}.{date}",
                           "{volname}.{basename}.{volname}",
                           "{volname}.{date}.{date}", "{volname}.{basename}",
                           "{volname}.{date}", "{volname}.{volname}"]

    def setUp(self):
        self.test_log = logbook.TestHandler()
        self.test_log.push_thread()
        snaprotator.logsetup._has_warnings = 0

    @contextmanager
    def assert_exits_with_error(self, re_str, exit_code):
        try:
            yield
        except SystemExit as e:
            self.assertTrue(self.test_log.has_error(
                re.compile(re_str)), self.test_log.formatted_records)
            self.assertEqual(exit_code, e.code)
        else:
            self.fail(
                "expected test to fail with an error like {!r}".format(re_str))

    def tearDown(self):
        self.test_log.pop_thread()

    @classmethod
    def _setup_sim(cls):
        enable()
        cls.simulator = Simulator()
        cls.simulator.activate()
        cls.simulator.set_version('5.0.0')
        cls.simulator.set_serial(4096)
        cls.mock_infinibox = InfiniBox(
            cls.simulator, auth=('infinidat', '123456'))
        cls.mock_infinibox.login()

    @classmethod
    def _setup_pool(cls):
        Pool = cls.mock_infinibox.objects.pools
        cls.pool = Pool.create(name='testpool', virtual_capacity=2 * TiB,
                               physical_capacity=2 * TiB)

    @classmethod
    def _setup_volumes(cls):
        volnames = ["{}_{}".format(cls.VOL_PREFIX, i)
                    for i in range(0, cls.MAXVOLS)]
        fsnames = ["{}_{}".format(cls.FS_PREFIX, i)
                   for i in range(0, cls.MAXVOLS)]
        for volname in volnames:
            vol = cls.mock_infinibox.objects.volumes.create(pool=cls.pool,
                                                            name=volname,
                                                            size=5 * GiB)
            snapnames = set(name_template.format(volname=volname,
                                                 basename=basename,
                                                 date=creation_date)
                            for name_template in cls.TEST_NAME_TEMPLATES
                            for creation_date in cls.CREATION_DATES
                            for basename in cls.BASENAMES)
            for snapname in sorted(snapnames):
                vol.create_snapshot(name=snapname)
        for fsname in fsnames:
            fs = cls.mock_infinibox.objects.filesystems.create(pool=cls.pool,
                                                               name=fsname,
                                                               size=5 * GiB)
            snapnames = set(name_template.format(volname=fsname,
                                                 basename=basename,
                                                 date=creation_date)
                            for name_template in cls.TEST_NAME_TEMPLATES
                            for creation_date in cls.CREATION_DATES
                            for basename in cls.BASENAMES)
            for snapname in sorted(snapnames):
                fs.create_snapshot(name=snapname)

    @classmethod
    def setup_class(cls):
        cls._setup_sim()
        cls._setup_pool()
        cls._setup_volumes()

    @classmethod
    def teardown_class(self):
        pass

    # def setup(self):
    #     self.log_handler = logbook.TestHandler()
    #     self.log_handler.push_thread()

    # def teardown(self):
    #     self.log_handler.pop_thread()

    def _create_vmware_config_file(self, infinibox_username='infinidat', infinibox_password='123456'):
        config = {
            'datastores': {'*': ''},
            'nomemory': {'*': '', '!*xxx*': ''},
            'logging': {'syslog': True, 'rotation': False, 'logpath': '/var/log/infinisnap/infinisnap.log', 'quiet': False},
            'infinibox': {'username': infinibox_username, 'host': 'ibox515', 'password': infinibox_password},
            'global': {'snap_basename': 'daily', 'snaps_to_keep': '7'},
            'noquiesce': {'*': '', '!xxx': ''},
            'vcenter': {'username': 'administrator', 'host': 'vsc-vc.lab.il.infinidat.com', 'password': 'vmwareesx'},
            'vms': {'*': ''}
        }
        return config

    def _create_simulator_config_file(self, *args, **kwargs):
        config = self._create_vmware_config_file(*args, **kwargs)
        config['infinibox']['host'] = self.simulator
        return config

    def _create_base_config_file(self):
        config = {
            'infinibox': {'username': 'infinidat', 'host': 'ibox515', 'password': '123456'},
            'global': {'snap_basename': 'daily', 'snaps_to_keep': '7'},
            'volumes': {'*': ''}
        }
        validate, out = snaprotator.config.validate_config(config)
        assert validate
        return snaprotator.utils.to_defaultdict(out)

    def get_infinidat_block_devices(self):
        devices = ['naa.6742b0f00000100000000000000003f0',
                   'naa.6742b0f0000010000000000000000426', 'naa.6742b0f000001000000000000000045c']
        res = list()
        for device in devices:
            dev = Mock(dev=device, name='/vmfs/devices/disks/{}'.format(device))
            dev.get_canonical_name.return_value = dev.dev
            res.append(dev)
        return res

    def _create_vm(self, name, datastore, template=False, power=True):
        def create_snapshot(snapshot, name, childSnapshotList, datastore):
            snap = Mock(
                snapshot=snapshot,
                name=name,
                _moId=snapshot,
                datastore=datastore,
                childSnapshotList=childSnapshotList
            )
            snap._moId = 'snap-' + name
            snap.__str__ = Mock()
            snap.__str__.return_value = 'snap-' + name
            return[snap]

        snap3 = create_snapshot(
            'vim.vm.Snapshot:snapshot-426', 'test3', [], datastore)
        snap2 = create_snapshot(
            'vim.vm.Snapshot:snapshot-425', 'test2', snap3, datastore)
        snap1 = create_snapshot(
            'vim.vm.Snapshot:snapshot-424', 'test1', snap2, datastore)
        snapshot = Mock(rootSnapshotList=snap1)
        if power:
            powerstate = 'poweredOn'
        else:
            powerstate = 'poweredOff'
        from pyVmomi import vim
        hardware = Mock()

        dev = MagicMock(spec=getattr(vim, 'VirtualDisk'))

        dev.backing = MagicMock(spec=getattr(
            vim, 'VirtualDiskFlatVer2BackingInfo'))

        # dev.backing = Mock()
        dev.backing.datastore = datastore
        hardware.device = [dev]
        vm = Mock(
            summary=Mock(
                config=Mock(
                    template=template),
                runtime=Mock(
                    powerState=powerstate)),
            name=name,
            datastore=datastore,
            snapshot=snapshot,
            config=Mock(hardware=hardware),
        )
        vm._moId = 'vm-' + name
        vm.name = name
        vm.__str__ = Mock()
        vm.__str__.return_value = "'vim.VirtualMachine:vm-{}'".format(
            hash(name) % 100000)

        return vm

    def _create_vms(self, datastores=None):
        names = ['vm1', 'vm2', 'vm3', 'vm4', 'vm5',
                 'vm6', 'vm7', 'vm8', 'template1', 'vm9-off', 'vm10-off']
        esx_host_names = ['esxhost1', 'esxhost2', 'esxhost3']
        if not datastores:
            datastores = self._create_datastores()
        esx_hosts = list()
        for index, host in enumerate(esx_host_names):
            vms = list()
            for name in names:
                datastore = datastores[index % 4]
                vm = self._create_vm(
                    host + '_' + 'ds{}_'.format((index % 4) + 1) + name, datastore, 'template' in name, 'off' not in name)
                vms.append(vm)
            esx_host = Mock(name=host, vm=vms, _moId='host-' + host)
            esx_hosts.append(esx_host)
        return esx_hosts

    def _create_vmfs_datastore(self, name, disknames):
        extent = [Mock(diskName=s) for s in disknames]
        disk = Mock(name=name)
        disk._moId = 'ds-' + name
        disk.name = name
        disk.summary = Mock(type='VMFS')
        disk.info = Mock(vmfs=Mock(extent=extent))
        disk.__str__ = Mock()

        disk.__str__.return_value = "'vim.Datastore:datastore-{}'".format(
            hash(name) % 100000)
        return disk

    def _create_nfs_datastore(self, name):
        disk = Mock(name=name)
        disk.name = name
        disk.summary = Mock(type='NFS')
        disk.info = Mock()
        return disk

    def _create_datastores(self):

        # datastore.summary.type
        # datastore.info.vmfs.extent
        ds1 = self._create_vmfs_datastore(
            'FirstDatastore', ['naa.6742b0f00000100000000000000003f0', 'naa.6742b0f0000010000000000000000426'])
        ds2 = self._create_vmfs_datastore(
            'Second Datastore with space', ['naa.6742b0f000001000000000000000045c'])
        ds3 = self._create_vmfs_datastore(
            'Third non infinidat datastore', ['naa.2020202020202020'])
        ds4 = self._create_nfs_datastore('NFS datastore')

        # caller = inspect.stack()[1][3]
        # if caller != '_create_vms':
        #     vms = self._create_vms()
        #     ds1.vm = vms[0].vm
        #     ds2.vm = vms[1].vm
        #     ds3.vm = vms[2].vm

        datastores = [ds1, ds2, ds3, ds4]
        return datastores

    def _mock_client(self):
        client = MagicMock(autospec=Client)
        datastores = self._create_datastores()
        vms = self._create_vms(datastores)
        datastores[0].vm = vms[0].vm
        datastores[1].vm = vms[1].vm
        client.get_datastores.return_value = datastores
        client.get_host_systems.return_value = vms
        client.service_content.about.fullName = 'Mock Vcenter 9.9.9'
        client.service_content.extensionManager.extensionList = [
            Mock(
                key='com.infinidat.powertools',
                server=[
                    Mock(url='http://powertools.infinidat.com/plugin/info')]
            )
        ]
        client.session_manager.currentSession.key = '523ff033-bcc9-cbc9-2886-7c539d140ece'

        return client

    def powertools_http(self, url, data):
        if isinstance(data, str):
            data = json.loads(data)
        if isinstance(data, dict):
            snapshots = data.get('infinibox_snapshots', [])

            for i in snapshots:
                snap_id = i.get('snapshot_id')
                guid = data.get('guid')
                _vms = [a.get('moref') for a in data.get('vms')]
                vm_refs = ",".join(_vms)
                esx_hosts = self._create_vms()
                vms = [a for esx in esx_hosts for a in esx.vm if str(a)[
                    5:-1] in _vms]
                datastore = vms[0].datastore.name
                vm_list = ",".join([a.name for a in vms])

                snapshot = next(
                    (a for a in self.mock_infinibox.objects.volumes.find(id=snap_id)), None)
                metadata = {'vm_list': vm_list,
                            'vm_refs': vm_refs,
                            'name': 'Registered Snapshot {}'.format(snap_id),
                            'original_datastore': datastore,
                            'host.created_by': 'VMPowerTools/5.2.1',
                            'snap_group_uuid': guid,
                            'description': '',
                            'first_vm_uuid': '42156466-a523-5a5e-5453-25b9133a094e',
                            'first_vm_snapshot_id': '{}'.format(snap_id)}
                for k, v in metadata.items():
                    snapshot.set_metadata(k, v)

        if url == 'http://powertools.infinidat.com/plugin/info' and data == {}:
            xml = b'<?xml version="1.0" encoding="UTF-8"?>\n<scriptConfiguration version="5.2.1.0">\n    <key>com.infinidat.powertools</key>\n    <multiVCsupported>false</multiVCsupported>\n    <supportNonSecureCommunication>true</supportNonSecureCommunication>\n    <description>Host PowerTools for VMware</description>\n    <name>Host PowerTools for VMware</name>\n    <vendor>INFINIDAT, Ltd</vendor>\n    \n        \n            <extension parent="InventoryView.Cluster" privilege="Host.Config.Storage">\n                \n                    <title locale="en">INFINIDAT</title>\n                \n                <url>http://172.20.79.148:80/plugin/redirect/cluster</url>\n            </extension>\n        \n    \n        \n            <extension parent="InventoryView.Datastore" privilege="Host.Config.Storage">\n                \n                    <title locale="en">INFINIDAT</title>\n                \n                <url>http://172.20.79.148:80/plugin/redirect/datastore</url>\n            </extension>\n        \n    \n        \n            <extension parent="InventoryView.VirtualMachineFolder" privilege="Host.Config.Storage">\n                \n                    <title locale="en">INFINIDAT</title>\n                \n                <url>http://172.20.79.148:80/plugin/redirect/folder</url>\n            </extension>\n        \n    \n        \n            <extension parent="InventoryView.HostSystem" privilege="Host.Config.Storage">\n                \n                    <title locale="en">INFINIDAT</title>\n                \n                <url>http://172.20.79.148:80/plugin/redirect/host</url>\n            </extension>\n        \n    \n        \n    \n        \n            <extension parent="InventoryView.VirtualMachine" privilege="Host.Config.Storage">\n                \n                    <title locale="en">INFINIDAT</title>\n                \n                <url>http://172.20.79.148:80/plugin/redirect/virtualmachine</url>\n            </extension>\n        \n    \n</scriptConfiguration>'
            return True, xml
        return True, '{"success": true}'

    def test_validate_arguments(self):
        config = {}
        snaprotator.config.update_arguments_to_config({}, config)
        validate, error = snaprotator.config.validate_config(config)

    def test_validate_arguments_wrong(self):
        config = {}
        snaprotator.config.update_arguments_to_config(
            {'--retain': '1'}, config)
        validate, error = snaprotator.config.validate_config(config)
        assert not validate

    def test_validate_arguments_right(self):
        config = self._create_vmware_config_file()
        validate, error = snaprotator.config.validate_config(config)
        assert validate

    def test_validate_arguments_snaps_to_keep_large(self):
        config = self._create_vmware_config_file()
        config['global']['snaps_to_keep'] = '1000'
        validate, error = snaprotator.config.validate_config(config)
        assert not validate

    def test_validate_arguments_snaps_to_keep_small(self):
        config = self._create_vmware_config_file()
        config['global']['snaps_to_keep'] = '0'
        validate, error = snaprotator.config.validate_config(config)
        assert not validate

    def test_validate_arguments_snaps_to_keep_not_digit(self):
        config = self._create_vmware_config_file()
        config['global']['snaps_to_keep'] = 'notdigit'
        validate, error = snaprotator.config.validate_config(config)
        assert not validate
        self.assertIn(
            "{'global': {'snaps_to_keep': ['Not a valid integer.']}}", error)

    def test_validate_arguments_retain_and_purge(self):
        config = self._create_vmware_config_file()
        config['global']['snaps_to_keep'] = 5
        config['infinibox']['purge'] = True
        validate, error = snaprotator.config.validate_config(config)
        assert not validate
        self.assertIn(
            '{\'global\': "purge and snaps_to_keep can\'t be specified together"}', error)

    def test_validate_arguments_no_retain_no_purge(self):
        config = self._create_vmware_config_file()
        del config['global']['snaps_to_keep']
        validate, error = snaprotator.config.validate_config(config)
        assert not validate
        self.assertIn(
            "{'global': 'One of purge/snaps_to_keep/list/listjson/delete should be specified'}", error)

    def test_validate_arguments_missing_section(self):
        config = self._create_vmware_config_file()
        del config['vcenter']
        validate, error = snaprotator.config.validate_config(config)
        assert not validate
        self.assertIn(
            '{\'vcenter\': "[vcenter] is required with fields \'vms\' ,\'noquiesce\',\'nomemory\' and \'datastores\'"}', error)

    def test_validate_arguments_masked_password(self):
        config = self._create_vmware_config_file()
        validate, out = snaprotator.config.validate_config(config)
        assert validate
        password = '{}'.format(out.get('infinibox').get('password'))
        vcenter_password = '{}'.format(out.get('vcenter').get('password'))
        assert password == vcenter_password == '********'

    def test_validate_arguments_email_invalid_email(self):
        config = self._create_vmware_config_file()
        config['logging'] = dict()
        config['logging']['email'] = True
        config['logging']['emailsender'] = 'user'
        validate, out = snaprotator.config.validate_config(config)
        assert not validate
        self.assertIn(
            "{'logging': {'emailsender': ['Not a valid email address.']}}", out)

    def test_validate_arguments_email_no_recipient(self):
        config = self._create_vmware_config_file()
        config['logging'] = dict()
        config['logging']['email'] = True
        validate, out = snaprotator.config.validate_config(config)
        assert not validate
        self.assertIn(
            "{'logging': {'email': 'emailsender,emailrcpt,emailserver are mandatory when email is enabled'}}", out)

    def test_validate_arguments_email_emailuser_no_emailpassword(self):
        config = self._create_vmware_config_file()
        config['logging'] = dict()
        config['logging']['email'] = True
        config['logging']['emailsender'] = 'user@yahoo.com'
        config['logging']['emailrcpt'] = 'user@yahoo.com'
        config['logging']['emailserver'] = 'email.server.com'
        config['logging']['emailuser'] = 'user@yahoo.com'

        validate, out = snaprotator.config.validate_config(config)
        assert not validate
        self.assertIn(
            "{'logging': {'email': 'emailuser and emailpassword shoud both be present or absent'}}", out)

    def test_validate_arguments_logs_to_keep_no_rotation(self):
        config = self._create_vmware_config_file()
        config['logging'] = dict()
        config['logging']['logs_to_keep'] = 5
        config['logging']['rotation'] = False
        validate, out = snaprotator.config.validate_config(config)
        assert not validate
        self.assertIn(
            "{'logging': {'logs_to_keep': 'logs_to_keep requires rotation to be set'}}", out)

    def test_validate_arguments_logs_to_keep_too_small(self):
        config = self._create_vmware_config_file()
        config['logging'] = dict()
        config['logging']['logs_to_keep'] = 0
        config['logging']['rotation'] = True
        validate, out = snaprotator.config.validate_config(config)
        assert not validate
        self.assertIn(
            "{'logging': {'logs_to_keep': ['Must be greater than or equal to 1 and less than or equal to 100.']}}", out)

    def test_validate_arguments_logs_to_keep_too_large(self):
        config = self._create_vmware_config_file()
        config['logging'] = dict()
        config['logging']['logs_to_keep'] = 101
        config['logging']['rotation'] = True
        validate, out = snaprotator.config.validate_config(config)
        assert not validate
        self.assertIn(
            "{'logging': {'logs_to_keep': ['Must be greater than or equal to 1 and less than or equal to 100.']}}", out)

    def test_validate_arguments_volumes_with_vmware(self):
        config = self._create_vmware_config_file()
        config['infinibox']['volumes'] = 'aaa'
        validate, out = snaprotator.config.validate_config(config)
        assert not validate
        self.assertIn(
            '[vcenter] section and --volumes/[volumes]/[fs] can\'t be used simultaneously."}', out)

    def test_validate_arguments_volumes_twice(self):
        config = self._create_vmware_config_file()
        config['infinibox']['volumes'] = 'aaa'
        del config['vcenter']
        del config['datastores']
        del config['noquiesce']
        del config['nomemory']
        del config['vms']
        config['volumes'] = dict()
        validate, out = snaprotator.config.validate_config(config)
        assert not validate
        self.assertIn(
            '"--volumes([infinibox]volumes) section and [volumes]/[fs] can\'t be used simultaneously."}', out)

    def test_datastores(self):
        client = self._mock_client()
        datastores = snaprotator.vmware.get_all_infinidat_datastores(client)
        assert len(datastores) == 2

    def test_get_datastores_on_block_storage(self):
        client = self._mock_client()
        all_datastores = client.get_datastores()
        vmfs_datastores = snaprotator.vmware.get_datastores_on_block_storage(
            all_datastores)
        assert len(vmfs_datastores) == 3

    def test_iter_all_vms(self):
        client = self._mock_client()
        vms = list(snaprotator.vmware.iter_all_vms(client))
        assert len(vms) == 24

    def test_list_all_snapshots(self):
        client = self._mock_client()
        vms = list(snaprotator.vmware.iter_all_vms(client))
        vmname, vmdatastores, vm = vms[0]
        snapshots = snaprotator.vmware.list_all_snapshots(vm)
        assert len(snapshots) == 3

    def test_get_datastores_to_snapshot_all(self):
        client = self._mock_client()
        datastores = snaprotator.vmware.get_all_infinidat_datastores(client)
        patterns = {'*': ''}
        result_datastores = snaprotator.vmware.get_datastores_to_snapshot(
            datastores, patterns)
        assert len(result_datastores) == 2

    def test_get_datastores_to_snapshot_specific(self):
        client = self._mock_client()
        datastores = snaprotator.vmware.get_all_infinidat_datastores(client)
        patterns = {'*second*': ''}
        result_datastores = snaprotator.vmware.get_datastores_to_snapshot(
            datastores, patterns)
        assert len(result_datastores) == 1

    def test_get_datastores_to_snapshot_exclude(self):
        client = self._mock_client()
        datastores = snaprotator.vmware.get_all_infinidat_datastores(client)
        patterns = {'*': '', '!*second*': ''}
        result_datastores = snaprotator.vmware.get_datastores_to_snapshot(
            datastores, patterns)
        assert len(result_datastores) == 1

    def test_get_datastores_to_snapshot_space(self):
        client = self._mock_client()
        datastores = snaprotator.vmware.get_all_infinidat_datastores(client)
        patterns = {'Second Datastore*': ''}
        result_datastores = snaprotator.vmware.get_datastores_to_snapshot(
            datastores, patterns)
        assert len(result_datastores) == 1

    def test_get_datastores_to_snapshot_zero(self):
        client = self._mock_client()
        datastores = snaprotator.vmware.get_all_infinidat_datastores(client)
        patterns = {'blahblah': ''}
        result_datastores = snaprotator.vmware.get_datastores_to_snapshot(
            datastores, patterns)
        assert len(result_datastores) == 0

    def test_connect_to_infinibox(self):
        config = self._create_simulator_config_file()
        ibox = snaprotator.infinibox.connect_to_infinibox(config, ssl=False)
        # ibox.login()
        assert ibox.login()

    def test_infinibox_validate_wrong_password(self):
        config = self._create_simulator_config_file(
            infinibox_username='blah', infinibox_password='blah')
        filelist = glob.glob("/tmp/snaprotator_cookie*")
        for filename in filelist:
            os.unlink(filename)
        with self.assert_exits_with_error('username/password error', exit_code=snaprotator.errors.EXIT_CODE_IBOX_LOGIN_FAILED):
            ibox = snaprotator.snaprotator.connect_to_infinibox(
                config, ssl=False)
            snaprotator.snaprotator.infinibox_validate(ibox)

    def test_infinibox_validate_wrong_state(self):
        config = self._create_vmware_config_file()
        config['infinibox']['host'] = Simulator()
        with self.assertRaises(SystemExit) as err:
            ibox = snaprotator.snaprotator.connect_to_infinibox(
                config, ssl=False)
            snaprotator.snaprotator.infinibox_validate(ibox)

    def test_infinibox_validate_too_many_snapshots(self):
        try:
            snaprotator.infinibox.MAX_SYSTEM_SNAPSHOTS = 10
            config = self._create_simulator_config_file()
            ibox = snaprotator.snaprotator.connect_to_infinibox(
                config, ssl=False)
            ibox.login()
            with self.assert_exits_with_error('Maximum number of snapshot',
                                              exit_code=snaprotator.errors.EXIT_CODE_IBOX_TOO_MANY_SNAPSHOTS_PER_SYSTEM):
                snaprotator.snaprotator.infinibox_validate(ibox)
        finally:
            snaprotator.infinibox.MAX_SYSTEM_SNAPSHOTS = 100000

    def test_infinibox_validate_wrong_version(self):
        sim = Simulator()
        sim.activate()
        sim.set_version('9.9.0')
        config = self._create_vmware_config_file()
        config['infinibox']['host'] = sim
        ibox = snaprotator.snaprotator.connect_to_infinibox(config, ssl=False)
        ibox.login()
        with self.assert_exits_with_error('not supported by this version',
                                          exit_code=snaprotator.errors.EXIT_CODE_IBOX_VERSION_UNSUPPORTED):
            snaprotator.snaprotator.infinibox_validate(ibox)

    def test_volume_validate_nonexisting(self):
        config = self._create_simulator_config_file()
        ibox = snaprotator.snaprotator.connect_to_infinibox(config, ssl=False)
        ibox.login()
        with self.assert_exits_with_error('Failed to find volume', exit_code=snaprotator.errors.EXIT_CODE_IBOX_VOLUME_NOT_FOUND):
            snaprotator.infinibox.volume_validate(
                ibox, 'non_existing_volume')

    def test_volume_validate_too_many_snapshots(self):
        config = self._create_simulator_config_file()
        ibox = snaprotator.snaprotator.connect_to_infinibox(config, ssl=False)
        ibox.login()
        try:
            snaprotator.infinibox.MAX_VOL_SNAPSHOTS = 10
            volume, snapshots, too_many_snaps = snaprotator.infinibox.volume_validate(
                ibox, 'testvol_0')
            self.assertTrue(too_many_snaps)
        finally:
            snaprotator.infinibox.MAX_VOL_SNAPSHOTS = 998

    def test_filter_irrelevant_snapshots(self):
        volume, snapshots, too_many_snaps = snaprotator.infinibox.volume_validate(
            self.mock_infinibox, 'testvol_0')
        for basename in self.BASENAMES:
            self.assertEqual(len(snaprotator.rotation.filter_irrelevant_snapshots(snapshots, 'testvol_0', basename)),
                             len(self.CREATION_DATES))

    def test_sort_and_filter_snapshots(self):
        volume, snapshots, too_many_snaps = snaprotator.infinibox.volume_validate(
            self.mock_infinibox, 'testvol_0')
        for basename in self.BASENAMES:
            sorted_snapshots = snaprotator.rotation.sort_and_filter_snapshots(
                snapshots, 'testvol_0', basename)
            for index, date in enumerate(sorted(self.CREATION_DATES)):
                self.assertRegex(
                    sorted_snapshots[index].get_name(), re.compile(date))

    def test_trim_number_of_snapshots(self):
        volname = 'testvol_0'
        config = self._create_base_config_file()
        for basename in self.BASENAMES:
            config.general.snap_basename = basename
            config.general.snaps_to_keep = 2
            volume, snapshots, too_many_snaps = snaprotator.infinibox.volume_validate(
                self.mock_infinibox, volname)
            snaprotator.rotation.trim_number_of_snapshots(
                self.mock_infinibox, config, snapshots, volname)
            volume, snapshots, too_many_snaps = snaprotator.infinibox.volume_validate(
                self.mock_infinibox, volname)
            sorted_snapshots = snaprotator.rotation.sort_and_filter_snapshots(
                snapshots, volname, basename)

            self.assertEqual(len(sorted_snapshots), 2)

    def test_trim_number_of_snapshots_retention_bigger_than_num_of_snaps(self):
        config = self._create_base_config_file()
        config.general.snaps_to_keep = 100
        volname = 'testvol_0'
        for basename in self.BASENAMES:
            config.general.snap_basename = basename
            volume, snapshots, too_many_snaps = snaprotator.infinibox.volume_validate(
                self.mock_infinibox, volname)
            snaprotator.rotation.trim_number_of_snapshots(
                self.mock_infinibox, config, snapshots, volname)
            volume, snapshots_after_trim, too_many_snaps = snaprotator.infinibox.volume_validate(
                self.mock_infinibox, volname)
            self.assertEqual(len(snapshots), len(snapshots_after_trim))

    def test_trim_number_of_snapshots_wrong_vol_and_basename(self):
        config = self._create_base_config_file()
        config.general.snap_basename = 'blah'
        config.general.snaps_to_keep = 2
        volume, snapshots, too_many_snaps = snaprotator.infinibox.volume_validate(
            self.mock_infinibox, 'testvol_0')
        snaprotator.rotation.trim_number_of_snapshots(
            self.mock_infinibox, config, snapshots, 'blah')
        self.assertFalse(self.test_log.has_info(
            re.compile('Deleted')), self.test_log.formatted_records)
        self.assertFalse(self.test_log.has_error(
            re.compile('Failed')), self.test_log.formatted_records)

    def test_trim_number_of_snapshots_cant_delete(self):
        config = self._create_base_config_file()
        config.general.snap_basename = basename = 'daily'
        config.general.snaps_to_keep = 2
        volname = 'testvol_3'
        volume, snapshots, too_many_snaps = snaprotator.infinibox.volume_validate(
            self.mock_infinibox, volname)
        test_host = self.mock_infinibox.hosts.create(name='delete_guardian')
        for snap in snapshots:
            test_host.map_volume(snap)
        snaprotator.rotation.trim_number_of_snapshots(
            self.mock_infinibox, config, snapshots, volname)
        volume, snapshots, too_many_snaps = snaprotator.infinibox.volume_validate(
            self.mock_infinibox, volname)
        sorted_snapshots = snaprotator.rotation.sort_and_filter_snapshots(
            snapshots, volname, basename)
        self.assertTrue(self.test_log.has_warning(
            re.compile('is mapped. Not deleting.')), self.test_log.formatted_records)
        self.assertEqual(len(sorted_snapshots), len(self.CREATION_DATES))

    def test_create_latest_snapshot(self):
        config = self._create_base_config_file()
        config.general.snap_basename = basename = 'daily'
        volname = 'testvol_4'
        volume, snapshots, too_many_snaps = snaprotator.infinibox.volume_validate(
            self.mock_infinibox, volname)
        sorted_snapshots = snaprotator.rotation.sort_and_filter_snapshots(
            snapshots, volname, basename)
        snaprotator.rotation.create_latest_snapshot(
            self.mock_infinibox, config, volume)
        volume, snapshots_after_create, too_many_snaps = snaprotator.infinibox.volume_validate(
            self.mock_infinibox, volname)
        sorted_snapshots_after_create = snaprotator.rotation.sort_and_filter_snapshots(
            snapshots_after_create, volname, basename)
        self.assertEqual(len(sorted_snapshots_after_create),
                         len(sorted_snapshots) + 1)

    def test_create_latest_snapshot_new_basename(self):
        config = self._create_base_config_file()
        config.general.snap_basename = basename = 'new_basename'
        volname = 'testvol_4'
        volume, snapshots, too_many_snaps = snaprotator.infinibox.volume_validate(
            self.mock_infinibox, volname)
        sorted_snapshots = snaprotator.rotation.sort_and_filter_snapshots(
            snapshots, volname, basename)
        snaprotator.rotation.create_latest_snapshot(
            self.mock_infinibox, config, volume)
        volume, snapshots_after_create, too_many_snaps = snaprotator.infinibox.volume_validate(
            self.mock_infinibox, volname)
        sorted_snapshots_after_create = snaprotator.rotation.sort_and_filter_snapshots(
            snapshots_after_create, volname, basename)
        self.assertEqual(len(sorted_snapshots_after_create),
                         len(sorted_snapshots) + 1)

    def test_create_latest_snapshot_creation_failed(self):
        config = self._create_base_config_file()
        config.general.snap_basename = 'new_basename'
        snaprotator.rotation.create_latest_snapshot(
            self.mock_infinibox, config, None)
        self.assertTrue(self.test_log.has_warning(
            re.compile('Failed')), self.test_log.formatted_records)

    def test_password_encryption(self):
        import uuid
        password = uuid.uuid4().hex
        encrypted = snaprotator.utils.mask(password)
        decrypted = snaprotator.utils.unmask(encrypted)
        self.assertEqual(password, decrypted)

    def test_no_arguments(self):
        sys.argv = ['snaprotator']
        with io.StringIO() as buf:
            with redirect_stdout(buf):
                with self.assertRaises(SystemExit) as cm:
                    snaprotator.snaprotator.main()
            out = buf.getvalue()
            self.assertTrue(out.startswith('Usage:'))
            self.assertIn('Encrypt cleartext password', out)

    def test_invalid_arguments(self):
        sys.argv = ['snaprotator', '-Dd']
        with io.StringIO() as buf:
            with redirect_stderr(buf):
                with self.assertRaises(SystemExit) as cm:
                    snaprotator.snaprotator.main()
            out = buf.getvalue()
            self.assertIn(
                "'infinibox': ['Missing data for required field.']", out, msg=out)
            self.assertIn(
                "'snap_basename': ['Missing data for required field.']", out, msg=out)

    def test_invalid_configfile(self):
        sys.argv = ['snaprotator', '-c', 'non_existing_config.conf']
        with io.StringIO() as buf:
            with redirect_stderr(buf):
                with self.assertRaises(SystemExit) as cm:
                    snaprotator.snaprotator.main()
            out = buf.getvalue()
            expected = "Config file path not found. Exiting. (non_existing_c" \
                "onfig.conf)\n"
            self.assertEqual(out, expected, msg=out)

    def test_full_run(self):
        vol_objects_count = len(self.mock_infinibox.volumes.to_list())
        fs_objects_count = len(self.mock_infinibox.filesystems.to_list())
        config = '''
        [global]
        snap_basename = kuku
        snaps_to_keep = 2
        use_volid = True
        [infinibox]
        host = {}
        username = infinidat
        password = ENC:1yH5iGO1nm0j5Xj0tyyGfnu4=
        [volumes]
        *
        [fs]
        *
        '''.format(self.simulator.get_hostname())
        with open('/tmp/config.conf', 'w') as f:
            f.write(config)
        with io.StringIO() as stdout, io.StringIO() as stderr:
            # First run, should have +5 snapshots for all volumes and fs
            with redirect_stderr(stderr), redirect_stdout(stdout):
                sys.argv = ['snaprotator', '-c', '/tmp/config.conf']
                with self.assertRaises(SystemExit) as cm:
                    snaprotator.snaprotator.main()
            _stdout = stdout.getvalue()
            _stderr = stderr.getvalue()

            self.assertEqual(
                vol_objects_count + 5, len(self.mock_infinibox.volumes.to_list()), msg=_stdout + _stderr)
            self.assertEqual(
                fs_objects_count + 5, len(self.mock_infinibox.filesystems.to_list()), msg=_stdout + _stderr)
            sleep(1)
            # Second run, should have +10 snapshots for all volumes and fs
            with redirect_stderr(stderr), redirect_stdout(stdout):
                sys.argv = ['snaprotator', '-c', '/tmp/config.conf']
                with self.assertRaises(SystemExit) as cm:
                    snaprotator.snaprotator.main()
            _stdout = stdout.getvalue()
            _stderr = stderr.getvalue()

            self.assertEqual(
                vol_objects_count + 10, len(self.mock_infinibox.volumes.to_list()), msg=_stdout + _stderr)
            self.assertEqual(
                fs_objects_count + 10, len(self.mock_infinibox.filesystems.to_list()), msg=_stdout + _stderr)
            sleep(1)
            # Third run, should have +10 snapshots for all volumes and fs
            with redirect_stderr(stderr), redirect_stdout(stdout):
                sys.argv = ['snaprotator', '-c', '/tmp/config.conf']
                with self.assertRaises(SystemExit) as cm:
                    snaprotator.snaprotator.main()
            _stdout = stdout.getvalue()
            _stderr = stderr.getvalue()

            self.assertEqual(
                vol_objects_count + 10, len(self.mock_infinibox.volumes.to_list()), msg=_stdout + _stderr)
            self.assertEqual(
                fs_objects_count + 10, len(self.mock_infinibox.filesystems.to_list()), msg=_stdout + _stderr)
            # Clean run, should have number as original
            with redirect_stderr(stderr), redirect_stdout(stdout):
                sys.argv = ['snaprotator', '-R', '-c', '/tmp/config.conf']
                with self.assertRaises(SystemExit) as cm:
                    snaprotator.snaprotator.main()
            _stdout = stdout.getvalue()
            _stderr = stderr.getvalue()
            self.assertEqual(vol_objects_count, len(
                self.mock_infinibox.volumes.to_list()), msg=_stdout + _stderr)
            self.assertEqual(fs_objects_count, len(
                self.mock_infinibox.filesystems.to_list()), msg=_stdout + _stderr)

    def test_active_active(self):
        with ecosystem.SimulationContext(isolated_env=True):
            systems = list()
            for _ in range(2):
                sim = Simulator()
                ecosystem.infinilab.client.get_mocked_infinilab().add_system(sim)
                sim.activate()
                ibox = InfiniBox(
                    sim, auth=('infinidat', '123456'))
                ibox.login()
                systems.append(ibox)
            primary, secondary = systems
            Pool = primary.objects.pools
            pool = Pool.create()
            volume = primary.objects.volumes.create(pool=pool,
                                                    name='aa_volume',
                                                    size=5 * GiB)

            link = primary.networking.ensure_rmr_link(secondary)
            link.set_witness_address('witness')
            primary.replicas.replicate_volume(
                volume, link=link, remote_pool=secondary.pools.create(), replication_type="ACTIVE_ACTIVE")

            with patch('snaprotator.snaprotator.connect_to_infinibox') as ibox, \
                    patch('snaprotator.infinibox.connect_to_infinibox') as remote_ibox:
                ibox.return_value = primary
                remote_ibox.return_value = secondary

                primary_vol_objects_count = len(primary.volumes.to_list())
                secondary_vol_objects_count = len(secondary.volumes.to_list())

                config = '''
                [global]
                snap_basename = kuku
                snaps_to_keep = 2
                use_volid = True
                snap_both_aa_peers = True
                [infinibox]
                host = ibox123
                username = infinidat
                password = ENC:1yH5iGO1nm0j5Xj0tyyGfnu4=
                [volumes]
                *
                '''
                with open('/tmp/config.conf', 'w') as f:
                    f.write(config)
                with io.StringIO() as stdout, io.StringIO() as stderr:
                    # First run, should have +1 snapshots for all volumes
                    with redirect_stderr(stderr), redirect_stdout(stdout):
                        sys.argv = ['snaprotator', '-c', '/tmp/config.conf']
                        with self.assertRaises(SystemExit) as cm:
                            snaprotator.snaprotator.main()
                    _stdout = stdout.getvalue()
                    _stderr = stderr.getvalue()

                    self.assertEqual(cm.exception.code, 0,
                                     msg=_stdout + _stderr)
                    self.assertEqual(
                        primary_vol_objects_count + 1, len(primary.volumes.to_list()), msg=_stdout + _stderr)
                    self.assertEqual(
                        secondary_vol_objects_count + 1, len(secondary.volumes.to_list()), msg=_stdout + _stderr)
                    sleep(1)
                    # Second run, should have +2 snapshots for all volumes
                    with redirect_stderr(stderr), redirect_stdout(stdout):
                        sys.argv = ['snaprotator', '-c', '/tmp/config.conf']
                        with self.assertRaises(SystemExit) as cm:
                            snaprotator.snaprotator.main()
                    _stdout = stdout.getvalue()
                    _stderr = stderr.getvalue()

                    self.assertEqual(cm.exception.code, 0,
                                     msg=_stdout + _stderr)
                    self.assertEqual(
                        primary_vol_objects_count + 2, len(primary.volumes.to_list()), msg=_stdout + _stderr)
                    self.assertEqual(
                        secondary_vol_objects_count + 2, len(secondary.volumes.to_list()), msg=_stdout + _stderr)
                    sleep(1)
                    # Third run, should have +2 snapshots for all volumes
                    with redirect_stderr(stderr), redirect_stdout(stdout):
                        sys.argv = ['snaprotator', '-c', '/tmp/config.conf']
                        with self.assertRaises(SystemExit) as cm:
                            snaprotator.snaprotator.main()
                    _stdout = stdout.getvalue()
                    _stderr = stderr.getvalue()

                    self.assertEqual(cm.exception.code, 0,
                                     msg=_stdout + _stderr)

                    self.assertEqual(
                        primary_vol_objects_count + 2, len(primary.volumes.to_list()), msg=_stdout + _stderr)
                    self.assertEqual(
                        secondary_vol_objects_count + 2, len(secondary.volumes.to_list()), msg=_stdout + _stderr)

                    # Clean run, should have number as original
                    with redirect_stderr(stderr), redirect_stdout(stdout):
                        sys.argv = ['snaprotator', '-R',
                                    '-c', '/tmp/config.conf']
                        with self.assertRaises(SystemExit) as cm:
                            snaprotator.snaprotator.main()
                    _stdout = stdout.getvalue()
                    _stderr = stderr.getvalue()

                    self.assertEqual(cm.exception.code, 0,
                                     msg=_stdout + _stderr)
                    self.assertEqual(
                        primary_vol_objects_count, len(primary.volumes.to_list()), msg=_stdout + _stderr)
                    self.assertEqual(
                        secondary_vol_objects_count, len(secondary.volumes.to_list()), msg=_stdout + _stderr)

    def test_active_active_cg(self):
        with ecosystem.SimulationContext(isolated_env=True):
            systems = list()
            for _ in range(2):
                sim = Simulator()
                ecosystem.infinilab.client.get_mocked_infinilab().add_system(sim)
                sim.activate()
                ibox = InfiniBox(
                    sim, auth=('infinidat', '123456'))
                ibox.login()
                systems.append(ibox)
            primary, secondary = systems
            Pool = primary.objects.pools
            pool = Pool.create()
            cg = primary.objects.cons_groups.create(name='cg1', pool=pool)
            volume1 = primary.objects.volumes.create(pool=pool,
                                                     name='aa_volume1',
                                                     size=5 * GiB)
            cg.add_member(volume1)
            volume2 = primary.objects.volumes.create(pool=pool,
                                                     name='aa_volume2',
                                                     size=5 * GiB)

            cg.add_member(volume2)

            link = primary.networking.ensure_rmr_link(secondary)
            link.set_witness_address('witness')
            primary.replicas.replicate_volume(
                cg, link=link, remote_pool=secondary.pools.create(), replication_type="ACTIVE_ACTIVE")

            with patch('snaprotator.snaprotator.connect_to_infinibox') as ibox, \
                    patch('snaprotator.infinibox.connect_to_infinibox') as remote_ibox:
                ibox.return_value = primary
                remote_ibox.return_value = secondary
                primary_vol_objects_count = len(primary.volumes.to_list())
                secondary_vol_objects_count = len(secondary.volumes.to_list())

                config = '''
                [global]
                snap_basename = kuku
                snaps_to_keep = 2
                use_volid = True
                snap_both_aa_peers = True
                [infinibox]
                host = ibox123
                username = infinidat
                password = ENC:1yH5iGO1nm0j5Xj0tyyGfnu4=
                [volumes]
                *
                '''
                with open('/tmp/config.conf', 'w') as f:
                    f.write(config)
                with io.StringIO() as stdout, io.StringIO() as stderr:
                    # First run, should have +2 snapshots for all volumes
                    with redirect_stderr(stderr), redirect_stdout(stdout):
                        sys.argv = ['snaprotator', '-c', '/tmp/config.conf']
                        with self.assertRaises(SystemExit) as cm:
                            snaprotator.snaprotator.main()
                    _stdout = stdout.getvalue()
                    _stderr = stderr.getvalue()

                    self.assertEqual(cm.exception.code, 0,
                                     msg=_stdout + _stderr)
                    self.assertEqual(
                        primary_vol_objects_count + 2, len(primary.volumes.to_list()), msg=_stdout + _stderr)
                    self.assertEqual(
                        secondary_vol_objects_count + 2, len(secondary.volumes.to_list()), msg=_stdout + _stderr)
                    sleep(1)
                    # Second run, should have +4 snapshots for all volumes
                    with redirect_stderr(stderr), redirect_stdout(stdout):
                        sys.argv = ['snaprotator', '-c', '/tmp/config.conf']
                        with self.assertRaises(SystemExit) as cm:
                            snaprotator.snaprotator.main()
                    _stdout = stdout.getvalue()
                    _stderr = stderr.getvalue()

                    self.assertEqual(cm.exception.code, 0,
                                     msg=_stdout + _stderr)
                    self.assertEqual(
                        primary_vol_objects_count + 4, len(primary.volumes.to_list()), msg=_stdout + _stderr)
                    self.assertEqual(
                        secondary_vol_objects_count + 4, len(secondary.volumes.to_list()), msg=_stdout + _stderr)
                    sleep(1)
                    # Third run, should have +2 snapshots for all volumes
                    with redirect_stderr(stderr), redirect_stdout(stdout):
                        sys.argv = ['snaprotator', '-c', '/tmp/config.conf']
                        with self.assertRaises(SystemExit) as cm:
                            snaprotator.snaprotator.main()
                    _stdout = stdout.getvalue()
                    _stderr = stderr.getvalue()

                    self.assertEqual(cm.exception.code, 0,
                                     msg=_stdout + _stderr)
                    self.assertEqual(
                        primary_vol_objects_count + 4, len(primary.volumes.to_list()), msg=_stdout + _stderr)
                    self.assertEqual(
                        secondary_vol_objects_count + 4, len(secondary.volumes.to_list()), msg=_stdout + _stderr)

                    # Clean run, should have number as original
                    with redirect_stderr(stderr), redirect_stdout(stdout):
                        sys.argv = ['snaprotator', '-R',
                                    '-c', '/tmp/config.conf']
                        with self.assertRaises(SystemExit) as cm:
                            snaprotator.snaprotator.main()
                    _stdout = stdout.getvalue()
                    _stderr = stderr.getvalue()

                    self.assertEqual(cm.exception.code, 0,
                                     msg=_stdout + _stderr)
                    self.assertEqual(
                        primary_vol_objects_count, len(primary.volumes.to_list()), msg=_stdout + _stderr)
                    self.assertEqual(
                        secondary_vol_objects_count, len(secondary.volumes.to_list()), msg=_stdout + _stderr)

    def test_active_active_failed_remote(self):
        with ecosystem.SimulationContext(isolated_env=True):
            systems = list()
            for _ in range(2):
                sim = Simulator()
                ecosystem.infinilab.client.get_mocked_infinilab().add_system(sim)
                sim.activate()
                ibox = InfiniBox(
                    sim, auth=('infinidat', '123456'))
                ibox.login()
                systems.append(ibox)
            sim = Simulator()
            remote_ibox = InfiniBox(
                sim, auth=('infinidat', '1234567'))
            primary, secondary = systems
            Pool = primary.objects.pools
            pool = Pool.create()
            volume = primary.objects.volumes.create(pool=pool,
                                                    name='bb_volume',
                                                    size=5 * GiB)

            link = primary.networking.ensure_rmr_link(secondary)
            link.set_witness_address('witness')
            primary.replicas.replicate_volume(
                volume, link=link, remote_pool=secondary.pools.create(), replication_type="ACTIVE_ACTIVE")

            with patch('snaprotator.snaprotator.connect_to_infinibox') as ibox, \
                    patch('snaprotator.infinibox.connect_to_infinibox') as remote_ibox:
                ibox.return_value = primary
                remote_ibox.return_value = remote_ibox
                primary_vol_objects_count = len(primary.volumes.to_list())
                secondary_vol_objects_count = len(secondary.volumes.to_list())

                config = '''
                [global]
                snap_basename = kuku
                snaps_to_keep = 2
                use_volid = True
                snap_both_aa_peers = True
                [infinibox]
                host = ibox123
                username = infinidat
                password = ENC:1yH5iGO1nm0j5Xj0tyyGfnu4=
                [volumes]
                *
                '''
                with open('/tmp/config.conf', 'w') as f:
                    f.write(config)
                with io.StringIO() as stdout, io.StringIO() as stderr:
                    # First run, should have +1 snapshots for all volumes
                    with redirect_stderr(stderr), redirect_stdout(stdout):
                        sys.argv = ['snaprotator', '-c', '/tmp/config.conf']
                        with self.assertRaises(SystemExit) as cm:
                            snaprotator.snaprotator.main()
                    _stdout = stdout.getvalue()
                    _stderr = stderr.getvalue()

                    self.assertEqual(cm.exception.code, 2,
                                     msg=_stdout + _stderr)
                    self.assertEqual(
                        primary_vol_objects_count + 1, len(primary.volumes.to_list()), msg=_stdout + _stderr)
                    self.assertEqual(
                        secondary_vol_objects_count, len(secondary.volumes.to_list()), msg=_stdout + _stderr)
                    sleep(1)
                    # Second run, should have +2 snapshots for all volumes
                    with redirect_stderr(stderr), redirect_stdout(stdout):
                        sys.argv = ['snaprotator', '-c', '/tmp/config.conf']
                        with self.assertRaises(SystemExit) as cm:
                            snaprotator.snaprotator.main()
                    _stdout = stdout.getvalue()
                    _stderr = stderr.getvalue()

                    self.assertEqual(cm.exception.code, 2,
                                     msg=_stdout + _stderr)
                    self.assertEqual(
                        primary_vol_objects_count + 2, len(primary.volumes.to_list()), msg=_stdout + _stderr)
                    self.assertEqual(
                        secondary_vol_objects_count, len(secondary.volumes.to_list()), msg=_stdout + _stderr)
                    sleep(1)
                    # Third run, should have +2 snapshots for all volumes
                    with redirect_stderr(stderr), redirect_stdout(stdout):
                        sys.argv = ['snaprotator', '-c', '/tmp/config.conf']
                        with self.assertRaises(SystemExit) as cm:
                            snaprotator.snaprotator.main()
                    _stdout = stdout.getvalue()
                    _stderr = stderr.getvalue()

                    self.assertEqual(cm.exception.code, 2,
                                     msg=_stdout + _stderr)
                    self.assertEqual(
                        primary_vol_objects_count + 2, len(primary.volumes.to_list()), msg=_stdout + _stderr)
                    self.assertEqual(
                        secondary_vol_objects_count, len(secondary.volumes.to_list()), msg=_stdout + _stderr)

                    # Clean run, should have number as original
                    with redirect_stderr(stderr), redirect_stdout(stdout):
                        sys.argv = ['snaprotator', '-R',
                                    '-c', '/tmp/config.conf']
                        with self.assertRaises(SystemExit) as cm:
                            snaprotator.snaprotator.main()
                    _stdout = stdout.getvalue()
                    _stderr = stderr.getvalue()

                    self.assertEqual(cm.exception.code, 2,
                                     msg=_stdout + _stderr)
                    self.assertEqual(
                        primary_vol_objects_count, len(primary.volumes.to_list()), msg=_stdout + _stderr)
                    self.assertEqual(
                        secondary_vol_objects_count, len(secondary.volumes.to_list()), msg=_stdout + _stderr)

    def test_full_run_vmware_connection_error(self):
        vol_objects_count = len(
            self.mock_infinibox.volumes.to_list())
        config = '''
        [global]
        snap_basename = kuku
        snaps_to_keep = 2
        use_volid = True
        [infinibox]
        host = {}
        username = infinidat
        password = ENC:1yH5iGO1nm0j5Xj0tyyGfnu4=
        [vcenter]
        host = aaaaaaaa
        username = bbb
        password = ccc
        [vms]
        *
        [noquiesce]
        *
        [nomemory]
        *
        [datastores]
        *
        '''.format(self.simulator.get_hostname())
        with open('/tmp/config.conf', 'w') as f:
            f.write(config)
        with io.StringIO() as stdout, io.StringIO() as stderr:
            # First run, should have +3 snapshots for all volumes
            with redirect_stderr(stderr), redirect_stdout(stdout):
                sys.argv = ['snaprotator', '-D',
                            '-c', '/tmp/config.conf']
                with self.assertRaises(SystemExit) as cm:
                    snaprotator.snaprotator.main()

            _stdout = stdout.getvalue()
            _stderr = stderr.getvalue()
            self.assertEqual(cm.exception.code, 40,
                             msg=_stdout + _stderr)
            self.assertEqual(
                vol_objects_count + 0, len(self.mock_infinibox.volumes.to_list()), msg=_stdout + _stderr)
            self.assertIn('Failed to connect to VCenter', _stderr + _stdout)

    def test_full_run_vmware(self):
        with patch('snaprotator.vmware.Client') as client, \
                patch('snaprotator.vmware.get_infinidat_block_devices') as devices, \
                patch('snaprotator.powertools.http') as http, \
                patch('snaprotator.vmware.get_task_status') as task_status:
            task_status.return_value = 'success', 'successfully finished'
            client.return_value = self._mock_client()
            devices.return_value = self.get_infinidat_block_devices()
            http.side_effect = self.powertools_http
            vol_objects_count = len(
                self.mock_infinibox.volumes.to_list())
            config = '''
            [global]
            snap_basename = fullrunvmware
            snaps_to_keep = 2
            use_volid = True
            [infinibox]
            host = {}
            username = infinidat
            password = ENC:1yH5iGO1nm0j5Xj0tyyGfnu4=
            [vcenter]
            host = aaa
            username = bbb
            password = ccc
            [vms]
            *
            [noquiesce]
            *
            [nomemory]
            *
            [datastores]
            *
            '''.format(self.simulator.get_hostname())
            with open('/tmp/config.conf', 'w') as f:
                f.write(config)
            with io.StringIO() as stdout, io.StringIO() as stderr:
                # First run, should have +3 snapshots for all volumes
                with redirect_stderr(stderr), redirect_stdout(stdout):
                    sys.argv = ['snaprotator',
                                '-c', '/tmp/config.conf']
                    with self.assertRaises(SystemExit) as cm:
                        snaprotator.snaprotator.main()
                _stdout = stdout.getvalue()
                _stderr = stderr.getvalue()
                expected = "Registered VMS: ['esxhost2_ds2_vm1', 'esxhost2_ds2_vm2', 'esxhost2_ds2_vm3', 'esxhost2_ds2_vm4', 'esxhost2_ds2_vm5', 'esxhost2_ds2_vm6', 'esxhost2_ds2_vm7', 'esxhost2_ds2_vm8', 'esxhost2_ds2_vm9-off', 'esxhost2_ds2_vm10-off']"
                self.assertIn(expected, _stdout + _stderr,
                              msg=_stdout + _stderr)
                self.assertEqual(
                    vol_objects_count + 3, len(self.mock_infinibox.volumes.to_list()), msg=_stdout + _stderr)
                sleep(1)
                # Second run, should have +6 snapshots for all volumes
                with redirect_stderr(stderr), redirect_stdout(stdout):
                    sys.argv = ['snaprotator',
                                '-c', '/tmp/config.conf']
                    with self.assertRaises(SystemExit) as cm:
                        snaprotator.snaprotator.main()
                _stdout = stdout.getvalue()
                _stderr = stderr.getvalue()
                self.assertIn(expected, _stdout + _stderr,
                              msg=_stdout + _stderr)

                self.assertEqual(
                    vol_objects_count + 6, len(self.mock_infinibox.volumes.to_list()), msg=_stdout + _stderr)
                sleep(1)
                # Third run, should have +6 snapshots for all volumes
                with redirect_stderr(stderr), redirect_stdout(stdout):
                    sys.argv = ['snaprotator',
                                '-c', '/tmp/config.conf']
                    with self.assertRaises(SystemExit) as cm:
                        snaprotator.snaprotator.main()

                _stdout = stdout.getvalue()
                _stderr = stderr.getvalue()
                self.assertIn(expected, _stdout + _stderr,
                              msg=_stdout + _stderr)

                self.assertEqual(
                    vol_objects_count + 6, len(self.mock_infinibox.volumes.to_list()), msg=_stdout + _stderr)
                # Clean run, should have number as original
            with io.StringIO() as stdout, io.StringIO() as stderr:
                with redirect_stderr(stderr), redirect_stdout(stdout):
                    sys.argv = ['snaprotator', '-R',
                                '-c', '/tmp/config.conf']
                    with self.assertRaises(SystemExit) as cm:
                        snaprotator.snaprotator.main()
                _stdout = stdout.getvalue()
                _stderr = stderr.getvalue()
                removed = [a for a in _stderr.split(
                    '\n') if 'Successfuly removed snapshot (Registered Snapshot ' in a]
                self.assertEqual(len(removed), 6, msg=_stderr + _stdout)

                self.assertEqual(vol_objects_count, len(
                    self.mock_infinibox.volumes.to_list()), msg=_stdout + _stderr)

    def test_full_run_vmware_single_datastore(self):
        with patch('snaprotator.vmware.Client') as client, \
                patch('snaprotator.vmware.get_infinidat_block_devices') as devices, \
                patch('snaprotator.powertools.http') as http, \
                patch('snaprotator.vmware.get_task_status') as task_status:
            task_status.return_value = 'success', 'successfully finished'
            client.return_value = self._mock_client()
            devices.return_value = self.get_infinidat_block_devices()
            http.side_effect = self.powertools_http
            vol_objects_count = len(
                self.mock_infinibox.volumes.to_list())
            config = '''
            [global]
            snap_basename = vmfull
            snaps_to_keep = 2
            use_volid = True
            [infinibox]
            host = {}
            username = infinidat
            password = ENC:1yH5iGO1nm0j5Xj0tyyGfnu4=
            [vcenter]
            host = aaa
            username = bbb
            password = ccc
            [vms]
            *
            [noquiesce]
            *
            [nomemory]
            *
            [datastores]
            Second*
            '''.format(self.simulator.get_hostname())
            with open('/tmp/config.conf', 'w') as f:
                f.write(config)
            with io.StringIO() as stdout, io.StringIO() as stderr:
                # First run, should have +1 snapshots for all volumes
                with redirect_stderr(stderr), redirect_stdout(stdout):
                    sys.argv = ['snaprotator', '-D',
                                '-c', '/tmp/config.conf']
                    with self.assertRaises(SystemExit) as cm:
                        snaprotator.snaprotator.main()
                _stdout = stdout.getvalue()
                _stderr = stderr.getvalue()
                self.assertTrue(client.called, msg=_stdout + _stderr)
                self.assertEqual(
                    vol_objects_count + 1, len(self.mock_infinibox.volumes.to_list()), msg=_stdout + _stderr)
                sleep(1)
                # Second run, should have +2 snapshots for all volumes
                with redirect_stderr(stderr), redirect_stdout(stdout):
                    sys.argv = ['snaprotator',
                                '-c', '/tmp/config.conf']
                    with self.assertRaises(SystemExit) as cm:
                        snaprotator.snaprotator.main()
                _stdout = stdout.getvalue()
                _stderr = stderr.getvalue()

                self.assertEqual(
                    vol_objects_count + 2, len(self.mock_infinibox.volumes.to_list()), msg=_stdout + _stderr)
                sleep(1)
                # Third run, should have +2 snapshots for all volumes
                with redirect_stderr(stderr), redirect_stdout(stdout):
                    sys.argv = ['snaprotator',
                                '-c', '/tmp/config.conf']
                    with self.assertRaises(SystemExit) as cm:
                        snaprotator.snaprotator.main()

                _stdout = stdout.getvalue()
                _stderr = stderr.getvalue()

                self.assertEqual(
                    vol_objects_count + 2, len(self.mock_infinibox.volumes.to_list()), msg=_stdout + _stderr)
            with io.StringIO() as stdout, io.StringIO() as stderr:
                # Clean run, should have number as original
                with redirect_stderr(stderr), redirect_stdout(stdout):
                    sys.argv = ['snaprotator', '-R',
                                '-c', '/tmp/config.conf']
                    with self.assertRaises(SystemExit) as cm:
                        snaprotator.snaprotator.main()
                _stdout = stdout.getvalue()
                _stderr = stderr.getvalue()
                self.assertIn('ds: Second Datastore with space',
                              _stderr, msg=_stdout + _stderr)
                self.assertEqual(vol_objects_count, len(
                    self.mock_infinibox.volumes.to_list()), msg=_stdout + _stderr)

    def test_full_run_vmware_single_datastore_no_powertools(self):
        with patch('snaprotator.vmware.Client') as client, \
                patch('snaprotator.vmware.get_infinidat_block_devices') as devices, \
                patch('snaprotator.vmware.get_task_status') as task_status:
            task_status.return_value = 'success', 'successfully finished'
            client.return_value = self._mock_client()
            devices.return_value = self.get_infinidat_block_devices()
            vol_objects_count = len(
                self.mock_infinibox.volumes.to_list())
            config = '''
            [global]
            snap_basename = vmnopower
            snaps_to_keep = 2
            use_volid = True
            [infinibox]
            host = {}
            username = infinidat
            password = ENC:1yH5iGO1nm0j5Xj0tyyGfnu4=
            [vcenter]
            host = aaa
            username = bbb
            password = ccc
            [vms]
            *
            [noquiesce]
            *
            [nomemory]
            *
            [datastores]
            Second*
            '''.format(self.simulator.get_hostname())
            with open('/tmp/config.conf', 'w') as f:
                f.write(config)
            with io.StringIO() as stdout, io.StringIO() as stderr:
                # First run, should have +1 snapshots for all volumes
                with redirect_stderr(stderr), redirect_stdout(stdout):
                    sys.argv = ['snaprotator', '-D',
                                '-c', '/tmp/config.conf']
                    with self.assertRaises(SystemExit) as cm:
                        snaprotator.snaprotator.main()
                _stdout = stdout.getvalue()
                _stderr = stderr.getvalue()
                self.assertTrue(client.called, msg=_stdout + _stderr)
                self.assertEqual(
                    vol_objects_count + 1, len(self.mock_infinibox.volumes.to_list()), msg=_stdout + _stderr)
                sleep(1)
                # Second run, should have +2 snapshots for all volumes
                with redirect_stderr(stderr), redirect_stdout(stdout):
                    sys.argv = ['snaprotator',
                                '-c', '/tmp/config.conf']
                    with self.assertRaises(SystemExit) as cm:
                        snaprotator.snaprotator.main()
                _stdout = stdout.getvalue()
                _stderr = stderr.getvalue()

                self.assertEqual(
                    vol_objects_count + 2, len(self.mock_infinibox.volumes.to_list()), msg=_stdout + _stderr)
                sleep(1)
                # Third run, should have +2 snapshots for all volumes
                with redirect_stderr(stderr), redirect_stdout(stdout):
                    sys.argv = ['snaprotator',
                                '-c', '/tmp/config.conf']
                    with self.assertRaises(SystemExit) as cm:
                        snaprotator.snaprotator.main()

                _stdout = stdout.getvalue()
                _stderr = stderr.getvalue()

                self.assertEqual(
                    vol_objects_count + 2, len(self.mock_infinibox.volumes.to_list()), msg=_stdout + _stderr)
            with io.StringIO() as stdout, io.StringIO() as stderr:
                # Clean run, should have number as original
                with redirect_stderr(stderr), redirect_stdout(stdout):
                    sys.argv = ['snaprotator', '-R',
                                '-c', '/tmp/config.conf']
                    with self.assertRaises(SystemExit) as cm:
                        snaprotator.snaprotator.main()
                _stdout = stdout.getvalue()
                _stderr = stderr.getvalue()
                self.assertNotIn('ds: Second Datastore with space',
                                 _stderr, msg=_stdout + _stderr)
                self.assertIn(
                    'No Host Power Tools For VMWare found. Check appliance installed/powered on.', _stderr, msg=_stdout + _stderr)
                self.assertEqual(vol_objects_count, len(
                    self.mock_infinibox.volumes.to_list()), msg=_stdout + _stderr)

    def test_full_run_vmware_single_vm(self):
        with patch('snaprotator.vmware.Client') as client, \
                patch('snaprotator.vmware.get_infinidat_block_devices') as devices, \
                patch('snaprotator.powertools.http') as http, \
                patch('snaprotator.vmware.get_task_status') as task_status:
            task_status.return_value = 'success', 'successfully finished'
            client.return_value = self._mock_client()
            devices.return_value = self.get_infinidat_block_devices()
            http.side_effect = self.powertools_http
            vol_objects_count = len(
                self.mock_infinibox.volumes.to_list())
            config = '''
            [global]
            snap_basename = vmsingle
            snaps_to_keep = 2
            use_volid = True
            [infinibox]
            host = {}
            username = infinidat
            password = ENC:1yH5iGO1nm0j5Xj0tyyGfnu4=
            [vcenter]
            host = aaa
            username = bbb
            password = ccc
            [vms]
            *esxhost2_ds2_vm6*
            [noquiesce]
            *
            [nomemory]
            *
            [datastores]
            *Second*
            '''.format(self.simulator.get_hostname())
            with open('/tmp/config.conf', 'w') as f:
                f.write(config)
            with io.StringIO() as stdout, io.StringIO() as stderr:
                # First run, should have +1 snapshots for all volumes
                with redirect_stderr(stderr), redirect_stdout(stdout):
                    sys.argv = ['snaprotator', '-D',
                                '-c', '/tmp/config.conf']
                    with self.assertRaises(SystemExit) as cm:
                        snaprotator.snaprotator.main()
                _stdout = stdout.getvalue()
                _stderr = stderr.getvalue()
                self.assertTrue(client.called, msg=_stdout + _stderr)
                self.assertEqual(
                    vol_objects_count + 1, len(self.mock_infinibox.volumes.to_list()), msg=_stdout + _stderr)
                self.assertIn(
                    "VMS to snap: ['esxhost2_ds2_vm6']", _stdout + _stderr)
                sleep(1)
                # Second run, should have +2 snapshots for all volumes
                with redirect_stderr(stderr), redirect_stdout(stdout):
                    sys.argv = ['snaprotator',
                                '-c', '/tmp/config.conf']
                    with self.assertRaises(SystemExit) as cm:
                        snaprotator.snaprotator.main()
                _stdout = stdout.getvalue()
                _stderr = stderr.getvalue()

                self.assertEqual(
                    vol_objects_count + 2, len(self.mock_infinibox.volumes.to_list()), msg=_stdout + _stderr)
                sleep(1)
                # Third run, should have +2 snapshots for all volumes
                with redirect_stderr(stderr), redirect_stdout(stdout):
                    sys.argv = ['snaprotator',
                                '-c', '/tmp/config.conf']
                    with self.assertRaises(SystemExit) as cm:
                        snaprotator.snaprotator.main()

                _stdout = stdout.getvalue()
                _stderr = stderr.getvalue()

                self.assertEqual(
                    vol_objects_count + 2, len(self.mock_infinibox.volumes.to_list()), msg=_stdout + _stderr)
                # Clean run, should have number as original
                with redirect_stderr(stderr), redirect_stdout(stdout):
                    sys.argv = ['snaprotator', '-R',
                                '-c', '/tmp/config.conf']
                    with self.assertRaises(SystemExit) as cm:
                        snaprotator.snaprotator.main()
                _stdout = stdout.getvalue()
                _stderr = stderr.getvalue()
                self.assertEqual(vol_objects_count, len(
                    self.mock_infinibox.volumes.to_list()), msg=_stdout + _stderr)

    def test_full_run_vmware_exclude_datastore(self):
        with patch('snaprotator.vmware.Client') as client, \
                patch('snaprotator.vmware.get_infinidat_block_devices') as devices, \
                patch('snaprotator.powertools.http') as http, \
                patch('snaprotator.vmware.get_task_status') as task_status:
            task_status.return_value = 'success', 'successfully finished'
            client.return_value = self._mock_client()
            devices.return_value = self.get_infinidat_block_devices()
            http.side_effect = self.powertools_http
            vol_objects_count = len(
                self.mock_infinibox.volumes.to_list())
            config = '''
            [global]
            snap_basename = vmexclude
            snaps_to_keep = 2
            use_volid = True
            [infinibox]
            host = {}
            username = infinidat
            password = ENC:1yH5iGO1nm0j5Xj0tyyGfnu4=
            [vcenter]
            host = aaa
            username = bbb
            password = ccc
            [vms]
            *
            [noquiesce]
            *
            [nomemory]
            *
            [datastores]
            *
            !First*
            '''.format(self.simulator.get_hostname())
            with open('/tmp/config.conf', 'w') as f:
                f.write(config)
            with io.StringIO() as stdout, io.StringIO() as stderr:
                # First run, should have +1 snapshots for all volumes
                with redirect_stderr(stderr), redirect_stdout(stdout):
                    sys.argv = ['snaprotator', '-D',
                                '-c', '/tmp/config.conf']
                    with self.assertRaises(SystemExit) as cm:
                        snaprotator.snaprotator.main()
                _stdout = stdout.getvalue()
                _stderr = stderr.getvalue()
                self.assertTrue(client.called, msg=_stdout + _stderr)
                self.assertEqual(
                    vol_objects_count + 1, len(self.mock_infinibox.volumes.to_list()), msg=_stdout + _stderr)
                sleep(1)
                # Second run, should have +2 snapshots for all volumes
                with redirect_stderr(stderr), redirect_stdout(stdout):
                    sys.argv = ['snaprotator',
                                '-c', '/tmp/config.conf']
                    with self.assertRaises(SystemExit) as cm:
                        snaprotator.snaprotator.main()
                _stdout = stdout.getvalue()
                _stderr = stderr.getvalue()

                self.assertEqual(
                    vol_objects_count + 2, len(self.mock_infinibox.volumes.to_list()), msg=_stdout + _stderr)
                sleep(1)
                # Third run, should have +2 snapshots for all volumes
                with redirect_stderr(stderr), redirect_stdout(stdout):
                    sys.argv = ['snaprotator',
                                '-c', '/tmp/config.conf']
                    with self.assertRaises(SystemExit) as cm:
                        snaprotator.snaprotator.main()

                _stdout = stdout.getvalue()
                _stderr = stderr.getvalue()

                self.assertEqual(
                    vol_objects_count + 2, len(self.mock_infinibox.volumes.to_list()), msg=_stdout + _stderr)
                # Clean run, should have number as original
                with redirect_stderr(stderr), redirect_stdout(stdout):
                    sys.argv = ['snaprotator', '-R',
                                '-c', '/tmp/config.conf']
                    with self.assertRaises(SystemExit) as cm:
                        snaprotator.snaprotator.main()
                _stdout = stdout.getvalue()
                _stderr = stderr.getvalue()
                self.assertEqual(vol_objects_count, len(
                    self.mock_infinibox.volumes.to_list()), msg=_stdout + _stderr)

    def test_full_run_invalid_password(self):
        config = '''
        [global]
        snap_basename = kuku
        snaps_to_keep = 2
        use_volid = True
        [infinibox]
        host = ibox123
        username = infinidat
        password = ENC:1yH5iGO1nm0j5Xj0tyyGfnu4
        [volumes]
        *
        [fs]
        *
        '''
        with open('/tmp/config.conf', 'w') as f:
            f.write(config)
        with io.StringIO() as stdout, io.StringIO() as stderr:
            # First run, should have +5 snapshots for all volumes and fs
            with redirect_stderr(stderr), redirect_stdout(stdout):
                sys.argv = ['snaprotator', '-c', '/tmp/config.conf']
                with self.assertRaises(SystemExit) as cm:
                    snaprotator.snaprotator.main()
            _stdout = stdout.getvalue()
            _stderr = stderr.getvalue()

            self.assertIn(
                "{'infinibox': {'password': ['Incorrect padding']}}", _stderr)

    def test_full_run_998_snaps(self):
        sim = Simulator()
        sim.activate()
        ibox = InfiniBox(sim, auth=('infinidat', '123456'))
        ibox.login()
        config = '''
        [global]
        snap_basename = max
        snaps_to_keep = 998
        use_volid = True
        [infinibox]
        host = {}
        username = infinidat
        password = ENC:1yH5iGO1nm0j5Xj0tyyGfnu4=
        [logging]
        logpath = /tmp/snap.log
        [volumes]
        max
        [fs]
        '''.format(sim.get_hostname())
        with open('/tmp/config.conf', 'w') as f:
            f.write(config)
        pool = ibox.pools.create()
        volume = ibox.volumes.create(name='max', pool=pool)
        vol_objects_count = len(ibox.volumes.to_list())
        for i in range(1, 995):
            name = '{}.max.1{:03d}.07.21_221105'.format('testvol_0', i)
            volume.create_snapshot(name)

        with io.StringIO() as stdout, io.StringIO() as stderr:
            with redirect_stderr(stderr), redirect_stdout(stdout):
                for i in range(1, 5):
                    start = time()
                    sys.argv = ['snaprotator', '-c', '/tmp/config.conf']
                    with self.assertRaises(SystemExit) as cm:
                        snaprotator.snaprotator.main()
                    _stdout = stdout.getvalue()
                    _stderr = stderr.getvalue()
                    self.assertEqual(cm.exception.code, 0,
                                     msg=_stdout + _stderr)
                    self.assertEqual(
                        vol_objects_count + 994 + i, len(ibox.volumes.to_list()), msg=_stdout + _stderr)
                    delay = 1 - (time() - start)
                    if delay > 0:
                        sleep(delay)

                for i in range(0, 5):
                    start = time()
                    sys.argv = ['snaprotator', '-c', '/tmp/config.conf']
                    with self.assertRaises(SystemExit) as cm:
                        snaprotator.snaprotator.main()
                    _stdout = stdout.getvalue()
                    _stderr = stderr.getvalue()
                    self.assertEqual(cm.exception.code, 0,
                                     msg=_stdout + _stderr)
                    self.assertEqual(
                        vol_objects_count + 998, len(ibox.volumes.to_list()), msg=_stdout + _stderr)
                    delay = 1 - (time() - start)
                    if delay > 0:
                        sleep(delay)
        volume.delete()

    def test_full_run_998_snaps_too_many(self):
        sim = Simulator()
        sim.activate()
        ibox = InfiniBox(sim, auth=('infinidat', '123456'))
        ibox.login()
        config = '''
        [global]
        snap_basename = max
        snaps_to_keep = 998
        use_volid = True
        [infinibox]
        host = {}
        username = infinidat
        password = ENC:1yH5iGO1nm0j5Xj0tyyGfnu4=
        [volumes]
        max
        [fs]
        '''.format(sim.get_hostname())
        with open('/tmp/config.conf', 'w') as f:
            f.write(config)
        pool = ibox.pools.create()
        volume = ibox.volumes.create(name='max', pool=pool)
        with io.StringIO() as stdout, io.StringIO() as stderr:
            with redirect_stderr(stderr), redirect_stdout(stdout):
                vol_objects_count = len(ibox.volumes.to_list())

                for i in range(1, 995):
                    name = '{}.max.1{:03d}.07.21_221105'.format('testvol_0', i)
                    volume.create_snapshot(name)
                volume.create_snapshot('aaa')
                volume.create_snapshot('bbb')
                volume.create_snapshot('ccc')

                for i in range(1, 3):
                    start = time()
                    sys.argv = ['snaprotator', '-c', '/tmp/config.conf']
                    with self.assertRaises(SystemExit) as cm:
                        snaprotator.snaprotator.main()
                    _stdout = stdout.getvalue()
                    _stderr = stderr.getvalue()
                    self.assertEqual(cm.exception.code, 0,
                                     msg=_stdout + _stderr)
                    self.assertEqual(
                        vol_objects_count + 997 + i, len(ibox.volumes.to_list()), msg=_stdout + _stderr)
                    delay = 1 - (time() - start)
                    if delay > 0:
                        sleep(delay)

                for i in range(1, 5):
                    start = time()
                    sys.argv = ['snaprotator', '-c', '/tmp/config.conf']
                    with self.assertRaises(SystemExit) as cm:
                        snaprotator.snaprotator.main()
                    _stdout = stdout.getvalue()
                    _stderr = stderr.getvalue()
                    self.assertEqual(cm.exception.code, 2,
                                     msg=_stdout + _stderr)
                    self.assertEqual(
                        vol_objects_count + 999, len(ibox.volumes.to_list()), msg=_stdout + _stderr)
                    self.assertIn(
                        'Too many snaps, deleting snap before creating', _stdout + _stderr)
                    self.assertIn(
                        'Will keep 996 (required = 998)', _stdout + _stderr)
                    delay = 1 - (time() - start)
                    if delay > 0:
                        sleep(delay)

        volume.delete()

    def test_full_run_998_cg_snaps(self):
        sim = Simulator()
        sim.activate()
        ibox = InfiniBox(sim, auth=('infinidat', '123456'))
        ibox.login()
        config = '''
        [global]
        snap_basename = max
        snaps_to_keep = 998
        use_volid = True
        [infinibox]
        host = {}
        username = infinidat
        password = ENC:1yH5iGO1nm0j5Xj0tyyGfnu4=
        [logging]
        logpath = /tmp/snap.log
        [volumes]
        max
        [fs]
        '''.format(sim.get_hostname())
        with open('/tmp/config.conf', 'w') as f:
            f.write(config)
        pool = ibox.pools.create()
        vol1 = ibox.volumes.create(name='vol1', pool=pool)
        vol2 = ibox.volumes.create(name='vol2', pool=pool)
        cg = ibox.cons_groups.create(name='max', pool=pool)
        cg.add_member(vol1)
        cg.add_member(vol2)

        vol_objects_count = len(ibox.volumes.to_list())
        for i in range(1, 995):
            name = '{}.max.1{:03d}.07.21_221105'.format('testvol_0', i)
            cg.create_snapshot(name)

        with io.StringIO() as stdout, io.StringIO() as stderr:
            with redirect_stderr(stderr), redirect_stdout(stdout):
                for i in range(1, 5):
                    start = time()
                    sys.argv = ['snaprotator', '-c', '/tmp/config.conf']
                    with self.assertRaises(SystemExit) as cm:
                        snaprotator.snaprotator.main()
                    _stdout = stdout.getvalue()
                    _stderr = stderr.getvalue()
                    self.assertEqual(cm.exception.code, 0,
                                     msg=_stdout + _stderr)
                    self.assertEqual(
                        vol_objects_count + 2 * 994 + 2 * i, len(ibox.volumes.to_list()), msg=_stdout + _stderr)
                    delay = 1 - (time() - start)
                    if delay > 0:
                        sleep(delay)

                for i in range(0, 5):
                    start = time()
                    sys.argv = ['snaprotator', '-c', '/tmp/config.conf']
                    with self.assertRaises(SystemExit) as cm:
                        snaprotator.snaprotator.main()
                    _stdout = stdout.getvalue()
                    _stderr = stderr.getvalue()
                    self.assertEqual(cm.exception.code, 0,
                                     msg=_stdout + _stderr)
                    self.assertEqual(
                        vol_objects_count + 2 * 998, len(ibox.volumes.to_list()), msg=_stdout + _stderr)
                    delay = 1 - (time() - start)
                    if delay > 0:
                        sleep(delay)
        cg.delete()

    def test_full_run_998_cg_snaps_too_many(self):
        sim = Simulator()
        sim.activate()
        ibox = InfiniBox(sim, auth=('infinidat', '123456'))
        ibox.login()
        config = '''
        [global]
        snap_basename = max
        snaps_to_keep = 998
        use_volid = True
        [infinibox]
        host = {}
        username = infinidat
        password = ENC:1yH5iGO1nm0j5Xj0tyyGfnu4=
        [volumes]
        max
        [fs]
        '''.format(sim.get_hostname())
        with open('/tmp/config.conf', 'w') as f:
            f.write(config)
        pool = ibox.pools.create()
        vol1 = ibox.volumes.create(name='vol1', pool=pool)
        vol2 = ibox.volumes.create(name='vol2', pool=pool)
        cg = ibox.cons_groups.create(name='max', pool=pool)
        cg.add_member(vol1)
        cg.add_member(vol2)

        with io.StringIO() as stdout, io.StringIO() as stderr:
            with redirect_stderr(stderr), redirect_stdout(stdout):
                vol_objects_count = len(ibox.volumes.to_list())

                for i in range(1, 995):
                    name = '{}.max.1{:03d}.07.21_221105'.format('testvol_0', i)
                    cg.create_snapshot(name)
                cg.create_snapshot('aaa')
                cg.create_snapshot('bbb')
                cg.create_snapshot('ccc')

                for i in range(1, 3):
                    start = time()
                    sys.argv = ['snaprotator', '-c', '/tmp/config.conf']
                    with self.assertRaises(SystemExit) as cm:
                        snaprotator.snaprotator.main()
                    _stdout = stdout.getvalue()
                    _stderr = stderr.getvalue()
                    self.assertEqual(cm.exception.code, 0,
                                     msg=_stdout + _stderr)
                    self.assertEqual(
                        vol_objects_count + 997 * 2 + i * 2, len(ibox.volumes.to_list()), msg=_stdout + _stderr)
                    delay = 1 - (time() - start)
                    if delay > 0:
                        sleep(delay)

                for i in range(1, 5):
                    start = time()
                    sys.argv = ['snaprotator', '-c', '/tmp/config.conf']
                    with self.assertRaises(SystemExit) as cm:
                        snaprotator.snaprotator.main()
                    _stdout = stdout.getvalue()
                    _stderr = stderr.getvalue()
                    self.assertEqual(cm.exception.code, 2,
                                     msg=_stdout + _stderr)
                    self.assertEqual(
                        vol_objects_count + 999 * 2, len(ibox.volumes.to_list()), msg=_stdout + _stderr)
                    self.assertIn(
                        'Too many snaps, deleting snap before creating', _stdout + _stderr)
                    self.assertIn(
                        'Will keep 996 (required = 998)', _stdout + _stderr)
                    delay = 1 - (time() - start)
                    if delay > 0:
                        sleep(delay)

        cg.delete()

    def test_full_retain_1(self):
        sim = Simulator()
        sim.activate()
        ibox = InfiniBox(sim, auth=('infinidat', '123456'))
        ibox.login()
        config = '''
        [global]
        snap_basename = one
        snaps_to_keep = 1
        use_volid = True
        [infinibox]
        host = {}
        username = infinidat
        password = ENC:1yH5iGO1nm0j5Xj0tyyGfnu4=
        [volumes]
        one
        [fs]
        '''.format(sim.get_hostname())
        with open('/tmp/config.conf', 'w') as f:
            f.write(config)
        pool = ibox.pools.create()
        vol1 = ibox.volumes.create(name='one', pool=pool)

        with io.StringIO() as stdout, io.StringIO() as stderr:
            with redirect_stderr(stderr), redirect_stdout(stdout):
                vol_objects_count = len(ibox.volumes.to_list())
                snapshots = vol1.get_snapshots().to_list()
                self.assertEqual(len(snapshots), 0)
                sys.argv = ['snaprotator', '-c', '/tmp/config.conf']
                with self.assertRaises(SystemExit) as cm:
                    snaprotator.snaprotator.main()
                _stdout = stdout.getvalue()
                _stderr = stderr.getvalue()
                self.assertEqual(cm.exception.code, 0,
                                     msg=_stdout + _stderr)
                snapshots = vol1.get_snapshots().to_list()
                self.assertEqual(len(snapshots), 1)
                sleep(1)
                for i in range(1, 10):
                    start = time()
                    snapshots = vol1.get_snapshots().to_list()
                    self.assertEqual(len(snapshots), 1)
                    with self.assertRaises(SystemExit) as cm:
                        snaprotator.snaprotator.main()
                    _stdout = stdout.getvalue()
                    _stderr = stderr.getvalue()
                    self.assertEqual(cm.exception.code, 0,
                                         msg=_stdout + _stderr)

                    snapshots = vol1.get_snapshots().to_list()
                    self.assertEqual(len(snapshots), 1, msg=f"i: {i} snaps {snapshots}")
                    delay = 1 - (time() - start)
                    if delay > 0:
                        sleep(delay)


    def test_full_retain_2(self):
        sim = Simulator()
        sim.activate()
        ibox = InfiniBox(sim, auth=('infinidat', '123456'))
        ibox.login()
        config = '''
        [global]
        snap_basename = two
        snaps_to_keep = 2
        use_volid = True
        [infinibox]
        host = {}
        username = infinidat
        password = ENC:1yH5iGO1nm0j5Xj0tyyGfnu4=
        [volumes]
        two
        [fs]
        '''.format(sim.get_hostname())
        with open('/tmp/config.conf', 'w') as f:
            f.write(config)
        pool = ibox.pools.create()
        vol1 = ibox.volumes.create(name='two', pool=pool)

        with io.StringIO() as stdout, io.StringIO() as stderr:
            with redirect_stderr(stderr), redirect_stdout(stdout):
                vol_objects_count = len(ibox.volumes.to_list())
                snapshots = vol1.get_snapshots().to_list()
                self.assertEqual(len(snapshots), 0)
                sys.argv = ['snaprotator', '-c', '/tmp/config.conf']
                with self.assertRaises(SystemExit) as cm:
                    snaprotator.snaprotator.main()
                _stdout = stdout.getvalue()
                _stderr = stderr.getvalue()
                self.assertEqual(cm.exception.code, 0,
                                     msg=_stdout + _stderr)
                snapshots = vol1.get_snapshots().to_list()
                self.assertEqual(len(snapshots), 1)
                sleep(1)
                with self.assertRaises(SystemExit) as cm:
                    snaprotator.snaprotator.main()
                _stdout = stdout.getvalue()
                _stderr = stderr.getvalue()
                self.assertEqual(cm.exception.code, 0,
                                     msg=_stdout + _stderr)
                snapshots = vol1.get_snapshots().to_list()
                self.assertEqual(len(snapshots), 2)
                sleep(1)
                for i in range(1, 10):
                    start = time()
                    snapshots = vol1.get_snapshots().to_list()
                    self.assertEqual(len(snapshots), 2)
                    with self.assertRaises(SystemExit) as cm:
                        snaprotator.snaprotator.main()
                    _stdout = stdout.getvalue()
                    _stderr = stderr.getvalue()
                    self.assertEqual(cm.exception.code, 0,
                                         msg=_stdout + _stderr)

                    snapshots = vol1.get_snapshots().to_list()
                    self.assertEqual(len(snapshots), 2, msg=f"i: {i} snaps {len(snapshots)}")
                    delay = 1 - (time() - start)
                    if delay > 0:
                        sleep(delay)