from time import time
from snaprotator.logsetup import UserLogger
from snaprotator import errors
import sys
import hashlib
import random
from base64 import b64encode, b64decode
from infi.pyutils.contexts import contextmanager
from types import SimpleNamespace
from collections import defaultdict

log = UserLogger(__name__)


class NestedDefaultDict(defaultdict):
    def __init__(self, *args, **kwargs):
        super(NestedDefaultDict, self).__init__(
            NestedDefaultDict, *args, **kwargs)

    def __repr__(self):
        return repr(dict(self)) if self else 'False'

    def __getattr__(self, name):
        return self[name]


def to_defaultdict(obj):
    if isinstance(obj, dict):
        return NestedDefaultDict({k.replace('global', 'general'): to_defaultdict(v) for k, v in obj.items()})
    elif isinstance(obj, list):
        return [to_defaultdict(a) for a in obj]
    else:
        return obj


def timed(func):
    def wrapper(*arg, **kw):
        t1 = time()
        res = func(*arg, **kw)
        elapsed = time() - t1
        log.debug("{} took {:.2f} seconds".format(func.__name__, elapsed))
        return res
    return wrapper


class dummy_context_mgr():
    def __enter__(self):
        return None

    def __exit__(self, exc_type, exc_value, traceback):
        return False


@contextmanager
def locking_context(name):
    from lockfile import AlreadyLocked
    from lockfile.pidlockfile import PIDLockFile
    from os import path
    lock = PIDLockFile(name, timeout=0)

    while True:
        try:
            if lock.i_am_locking():  # our PID is already acquired the lock
                log.uinfo(
                    "snaprotator process already holds the pid lock file")
                break
            lock.acquire(0)
            log.debug("Lock acquired successfully")
            break  # successfully acquired the lock
        except AlreadyLocked:
            locking_pid = lock.read_pid()  # either None or int
            if locking_pid is not None \
                    and path.exists("/proc/{}".format(locking_pid)):
                log.uabort_with_exc_info(
                    "another snaprotator process (pid {})"
                    " is still running and did not release the lock file"
                    .format(locking_pid),
                    exit_code=errors.EXIT_CODE_ALREADY_RUNNING)
            # pid file is left over by a dead process, break the lock and
            # re-acquire
            log.uinfo("detected lock left over of a dead snaprotator process"
                      " ({}), breaking lock".format(locking_pid))
            lock.break_lock()

    try:
        yield
    finally:
        lock.release()


def signal_handler(signal, frame):
    msg = ('Execution has been interrupted by Ctrl+C!\n'
           'Please re-run the tool, to make sure the rotation is in order'
           )
    log.uwarn(msg)
    sys.exit(errors.EXIT_CODE_INTERRUPT)


PY3 = sys.version_info[0] == 3
MAGIC_VER = 'ENC:1'

default_mask_key = hashlib.sha512(b'InF').digest()


def _ord(c):
    # single char (byte/str) to int
    if isinstance(c, int):
        return c
    return ord(c)


def _chr(o):
    # int to byte (if available) or str
    if PY3:
        return bytes([o])
    return chr(o)


def get_default_mask_key():
    return default_mask_key


def set_default_mask_key(k):
    global default_mask_key
    default_mask_key = k


def set_hash_seed(seed):
    set_default_mask_key(hashlib.sha512(seed).digest())


def mask(s, k=None):
    k = k or get_default_mask_key()
    s += '\x00' * ((16 - len(s) % 16) % 16)
    s1 = random.randint(1, 255)
    xored_s = b''.join([_chr(_ord(c) ^ (_ord(k[i % len(k)]) ^ s1))
                        for i, c in enumerate(s)])
    return MAGIC_VER + b64encode(_chr(s1) + xored_s).decode("ascii")


def unmask(s, k=None):
    k = k or get_default_mask_key()
    if not s.startswith(MAGIC_VER):
        return s
    s = b64decode(s[len(MAGIC_VER):])
    s1 = _ord(s[0])
    return ''.join([chr(_ord(c) ^ (_ord(k[i % len(k)]) ^ s1)) for i, c in
                    enumerate(s[1:])]).rstrip('\x00')
