import re
from time import sleep
from fnmatch import fnmatch
from snaprotator.logsetup import UserLogger
from pyVmomi import vim, vmodl
from infi.pyvmomi_wrapper import Client
from infi.storagemodel.vendor.infinidat.shortcuts import get_infinidat_block_devices
from requests.exceptions import (
    ConnectionError, ConnectTimeout, ReadTimeout, SSLError, ProxyError, RetryError)
from snaprotator import errors
from snaprotator.config import read_patterns_from_config
from snaprotator.utils import timed
from snaprotator.infinibox import get_all_volumes_by_serial, do_infinibox_snapshot
import snaprotator.powertools as powertools
import uuid
from functools import partial
from infi.storagemodel.vmware import with_host
from itertools import chain


log = UserLogger(__name__)


POWERTOOLS_PACKAGE_NAME = "com.infinidat.powertools"
DEFAULT_PARALLEL = 5


@timed
def connect_to_vcenter(config):
    hostname = config['vcenter']['host']
    username = config['vcenter']['username']
    password = config['vcenter']['password']

    log.uinfo("Connecting to VCenter '{}'".format(hostname))
    try:
        client = Client(hostname, username=username, password=password)
        log.uinfo("vCenter version: {}".format(
            client.service_content.about.fullName))
    except (ConnectionError, ConnectTimeout, ReadTimeout, SSLError, ProxyError, RetryError) as e:
        log.uabort_with_exc_info("VCenter connection error, aborting. {}", str(e),
                                 exit_code=errors.EXIT_CODE_VMWARE_NETWORK_ERROR)
    except vim.fault.InvalidLogin as e:
        log.uabort_with_exc_info("VCenter login error, aborting. {}", e.msg,
                                 exit_code=errors.EXIT_CODE_VMWARE_LOGIN_FAILED)
    except vim.fault.NoPermission as e:
        log.uabort_with_exc_info("VCenter login error. {}", e.msg,
                                 exit_code=errors.EXIT_CODE_VMWARE_LOGIN_FAILED)
    except SystemExit:
        raise
    except Exception as e:
        if hasattr(e, 'msg'):
            msg = e.msg
        else:
            msg = repr(e)
        log.uabort_with_exc_info("Failed to connect to VCenter, aborting. {}".format(msg),
                                 exit_code=errors.EXIT_CODE_VMWARE_GENERAL_ERROR)

    return client


def get_datastores_on_block_storage(datastores):
    return [datastore for datastore in datastores if datastore.summary.type == 'VMFS']


def is_infinibox_backed_vmfs_datastore(datastore):
    pred = partial(re.match, 'naa.6742b0f0')
    return any(pred(extent.diskName) for extent in datastore.info.vmfs.extent)


def get_all_infinidat_datastores(client):
    all_datastores = [a for a in client.get_datastores()
                      if a.summary.accessible]
    block_datastores = get_datastores_on_block_storage(all_datastores)
    return [datastore for datastore in block_datastores if is_infinibox_backed_vmfs_datastore(datastore)]


@timed
def iter_all_vms(client, show_powered_off=False, include_templates=False):
    log.info('include_templates {}'.format(include_templates))
    esx_hosts = client.get_host_systems()
    for esx_host in esx_hosts:
        for vm in esx_host.vm:
            # only add vm if not template and is powered on
            if include_templates or not vm.summary.config.template:
                if show_powered_off or vm.summary.runtime.powerState == 'poweredOn':
                    if vm is None or vm.config is None or vm.config.hardware is None:
                        continue
                    datastores = list(set(
                        dev.backing.datastore for dev in vm.config.hardware.device if isinstance(dev, vim.VirtualDisk)
                        and isinstance(dev.backing, vim.VirtualDiskFlatVer2BackingInfo)))
                    yield vm.name, ",".join(datastore.name for datastore in datastores), vm


def get_all_vms_on_datastore(datastore):
    return [vm for vm in datastore.vm]


def get_number_of_disk_datastores(vm):
    if vm is None or vm.config is None or vm.config.hardware is None:
        return 0
    return len(set(
        dev.backing.datastore for dev in vm.config.hardware.device if isinstance(dev, vim.VirtualDisk)
        and isinstance(dev.backing, vim.VirtualDiskFlatVer2BackingInfo)))


@timed
def get_all_vms_on_datastore_with_single_datastore(datastore, consistent_vms, consistent_only=True, include_templates=False):
    vm_dict = {vm: {"quiesce": quiesce, "memory": memory, "snapshot": True,
                    "moref": get_vm_string_name(vm)} for vm, quiesce, memory in consistent_vms}
    vms = list()
    vm_names = list()
    log.debug('Iterating VMs')
    for vm in datastore.vm:
        if get_number_of_disk_datastores(vm) == 1:
            vm_datastores = [dev.backing.datastore for dev
                             in vm.config.hardware.device
                             if isinstance(dev, vim.VirtualDisk)
                             and isinstance(dev.backing, vim.VirtualDiskFlatVer2BackingInfo)]
            vm_datastore = vm_datastores[0]
            log.debug('VM {} has 1 datastore {}'.format(vm.name, vm_datastore))
            if vm_datastore == datastore:
                if vm.summary.config.numVirtualDisks:
                    log.debug(' VM {} has disks.'.format(vm.name))
                    if vm in vm_dict:
                        log.debug(
                            'Include: VM {} is in consistent VM list'.format(vm.name))
                        vms.append(vm_dict[vm])
                        vm_names.append(vm.name)
                    else:
                        if vm.summary.config.template:
                            if include_templates:
                                log.debug(
                                    'Include: VM {} is a template and include_templates = True.'.format(vm.name))
                                vms.append({"quiesce": True, "memory": False,
                                            "snapshot": False, "moref": get_vm_string_name(vm)})
                                vm_names.append(vm.name)
                            else:
                                log.debug(
                                    'Exclude: VM {} is a template and include_templates = False.'.format(vm.name))

                        elif vm.summary.runtime.powerState == 'poweredOff':
                            log.debug(
                                'Include: VM {} is powered off'.format(vm.name))
                            vms.append({"quiesce": True, "memory": False,
                                        "snapshot": False, "moref": get_vm_string_name(vm)})
                            vm_names.append(vm.name)
                        elif consistent_only:
                            log.debug(
                                'Exclude: Not registering VM {}, it is powered on and consistent_only = True.'.format(vm.name))
                        else:  # Running VM, no VM snap, no quiesce, no memory
                            log.debug(
                                'Include: Registering VM {}, it is powered on and consistent_only = False'.format(vm.name))
                            vms.append({"quiesce": False, "memory": False,
                                        "snapshot": False, "moref": get_vm_string_name(vm)})
                            vm_names.append(vm.name)
                else:
                    log.debug(
                        'Exclude: Not registering VM {}, it has no disks'.format(vm.name))
            else:
                log.debug(
                    'Exclude: Not registering VM {}, it has no disks on this datastore'.format(vm.name))
        else:
            log.debug('Exclude: Not registering VM {}, it has {} datastores'.format(
                vm.name, get_number_of_disk_datastores(vm)))
    return vms, vm_names


def find_vmware_power_tools_url(client):
    try:
        e = [e for e in client.service_content.extensionManager.extensionList if e.key ==
             POWERTOOLS_PACKAGE_NAME][0]
        # http://1.2.3.4/something/somethingelse
        url = e.server[0].url
        log.debug('POWERTOOLS Url: {}'.format(url))
        # extract ip/name
        return url
    except Exception:
        return None


@timed
def iter_filtered_datastores_block_devices(client, filtered_datastores):
    def get_infinidat_devices_under_vmfs_datastore(datastore, infinidat_devs):
        underlying_devs = set(re.sub('^naa\.6', '', extent.diskName)
                              for extent in datastore.info.vmfs.extent)
        return underlying_devs & infinidat_devs

    esx_hosts = client.get_host_systems()
    for esx_host in esx_hosts:
        with with_host(client, esx_host):
            infinidat_devs = set(re.sub('^naa\.6', '', dev.get_canonical_name())
                                 for dev in get_infinidat_block_devices())
            datastores_to_remove = []
            for datastore in filtered_datastores:
                infinidat_devices = get_infinidat_devices_under_vmfs_datastore(
                    datastore, infinidat_devs)
                if infinidat_devices:
                    datastores_to_remove.append(datastore)
                    yield datastore, infinidat_devices
            for d in datastores_to_remove:
                filtered_datastores.remove(d)
        if not filtered_datastores:
            break


def initiate_snapshot_for_vm(client, snap_basename, vm, quiesce=True, memory=False, guid=None):
    vm_name = "<unknown>"
    task = None
    try:
        vm_name = vm.name
        log.uinfo('Creating snapshot for vm {} (quiesce = {} memory = {}) (guid = {})',
                  vm_name, quiesce, memory, guid)
        task = vm.CreateSnapshot_Task(
            name=snap_basename, memory=memory, quiesce=quiesce, description=guid)

    except vim.fault.NoPermission as e:
        log.uwarn("Snapshot creation failed for VM {} ({})".format(vm_name, e.msg))
    except Exception as e:
        if hasattr(e, 'msg'):
            msg = e.msg
        else:
            msg = repr(e)
        log.uwarn("Snapshot creation failed for VM {} {}".format(vm_name, msg))

    return task


def wait_for_result(task):
    while task.info.result is None:
        sleep(0.1)
    return task.info.result


def get_datastore_string_name(datastore):
    return str(datastore)[5:-1]


def get_vm_string_name(vm):
    return str(vm)[5:-1]


def initiate_remove_snapshot(vm_snapshot_pair):
    vm, snapshot = vm_snapshot_pair
    task = None
    try:
        log.uinfo('Initiate snapshot removal for VM {} ({})', vm.name, snapshot)
        task = snapshot.RemoveSnapshot_Task(removeChildren=False)
    except vim.ManagedObjectNotFound:
        log.uwarn(
            'The object has already been deleted or has not been completely created')
    except vmodl.fault.HostNotConnected as e:
        log.uwarn("Can't delete VM snapshot. {}", e.msg)
    except vim.fault.NoPermission as e:
        log.uwarn("Can't delete VM snapshot. ({})".format(e.msg))
    except Exception as e:
        if hasattr(e, 'msg'):
            msg = e.msg
        else:
            msg = repr(e)
        log.uwarn("Unknown VM snapshot remove error. {}".format(msg))
    return task


def remove_snapshot(client, snapshot):
    try:
        task = snapshot.RemoveSnapshot_Task(removeChildren=False)
        client.wait_for_task(task)
    except vim.ManagedObjectNotFound:
        log.uwarn(
            'The object has already been deleted or has not been completely created')
    except vmodl.fault.HostNotConnected as e:
        log.uwarn("Can't delete VM snapshot. {}", e.msg)
    except vim.fault.NoPermission as e:
        log.uwarn("Can't delete VM snapshot. ({})".format(e.msg))
    except Exception as e:
        if hasattr(e, 'msg'):
            msg = e.msg
        else:
            msg = repr(e)
        log.uwarn("Unknown VM snapshot remove error. {}".format(msg))


def list_all_snapshots(vm):
    def recursive_list_all_snapshots(snap, attr='childSnapshotList'):
        snap_list = getattr(snap, attr, [])
        return snap_list + list(chain(*[recursive_list_all_snapshots(o) for o in snap_list]))
    return recursive_list_all_snapshots(vm.snapshot, 'rootSnapshotList')


def cleanup_infinidat_snapshots(client, snap_basename, vm):
    snapshots = list(list_all_snapshots(vm))
    try:
        for snap in snapshots:
            log.debug('Snap name: {} Snap obj: {}', snap.name, snap.snapshot)
        for snapshot in snapshots:
            if snapshot.name == snap_basename:
                log.uinfo('Deleting snapshot {} for vm {}',
                          snapshot.name, vm.name)
                remove_snapshot(client, snapshot.snapshot)
    except Exception as e:
        log.uwarn('Unknown error in VMWare cleanup'.format(
            repr(e)), exc_info=True)


def get_task_status(task):
    message = None
    if task.info.state == 'error':
        if hasattr(task.info.error, 'reason'):
            message = task.info.error.reason
        elif hasattr(task.info.error, 'msg'):
            message = task.info.error.msg
        elif hasattr(task.info.error, 'message'):
            message = task.info.error.message
        else:
            message = 'Unknown'
    return task.info.state, message


def iter_filtered_datastores(datastores, pattern):
    for datastore in datastores:
        if fnmatch(datastore.name.lower(), pattern.lower()):
            yield datastore


def get_datastores_to_snapshot(datastores, patterns):
    def get_filtered_datastores(pred):
        result = set()
        for pattern in (p for p in patterns if pred(p)):
            if pattern[0] == '!':
                pattern = pattern[1:]
            _res = list(iter_filtered_datastores(datastores, pattern))
            if not _res:
                log.uwarn(
                    "Couldn't find datastores matching pattern {}".format(pattern))
            result.update(_res)
        return result

    includes = get_filtered_datastores(lambda p: p and p[0] != '!')
    excludes = get_filtered_datastores(lambda p: p and p[0] == '!')
    return list(includes - excludes)


def list_of_datastores_to_snapshot(config, datastores):
    patterns = read_patterns_from_config(config, 'datastores')
    return get_datastores_to_snapshot(datastores, patterns)


def iter_filtered_vms(pattern, datastores, vms):
    for vmname, vmdatastorename, vm in vms:
        for datastorename in vmdatastorename.split(","):
            if datastorename in datastores:
                if fnmatch(vmname.lower(), pattern.lower()):
                    yield vmname, vmdatastorename, vm


def get_vms_to_snapshot(patterns, datastores, vms, section):
    def get_filtered_vms(pred):
        result = set()
        for pattern in (p for p in patterns if pred(p)):
            if pattern[0] == '!':
                pattern = pattern[1:]
            _res = list(iter_filtered_vms(pattern, datastores, vms))
            if not _res and section:
                log.uwarn(
                    "Couldn't find vm matching pattern {} in section [{}]".format(pattern, section))
            result.update(_res)
        return result
    includes = get_filtered_vms(lambda p: p and p[0] != '!')
    excludes = get_filtered_vms(lambda p: p and p[0] == '!')
    return list(includes - excludes)


def list_of_vms_to_snapshot(config, datastores, vms, section=None):
    patterns = read_patterns_from_config(config, 'vms')
    return get_vms_to_snapshot(patterns, datastores, vms, section)


def list_of_vms_without_quiesce(config, datastores, vms, section=None):
    patterns = read_patterns_from_config(config, 'noquiesce')
    return get_vms_to_snapshot(patterns, datastores, vms, section)


def list_of_vms_without_memory(config, datastores, vms, section=None):
    patterns = read_patterns_from_config(config, 'nomemory')
    return get_vms_to_snapshot(patterns, datastores, vms, section)


def iter_vms(config, datastores, all_vms):
    vms = list_of_vms_to_snapshot(config, datastores, all_vms)
    noquiesce = list_of_vms_without_quiesce(config, datastores, vms)
    nomemory = list_of_vms_without_memory(config, datastores, vms)
    for vm in vms:
        quiesce = vm not in noquiesce
        memory = vm not in nomemory
        yield vm, quiesce, memory


def report_missing_vms(config, datastores, all_vms):
    vms = list_of_vms_to_snapshot(config, datastores, all_vms, section='vms')
    noquiesce = list_of_vms_without_quiesce(
        config, datastores, vms, section='noquiesce')
    nomemory = list_of_vms_without_memory(
        config, datastores, vms, section='nomemory')


def pipeline_vm_tasks(parallel, items, task_factory, task_handler):
    tasks = dict()
    while items or tasks:
        if len(tasks) < parallel and items:
            item = items.pop()
            task = task_factory(item)
            if task is not None:
                tasks[item] = task

        for item, task in tasks.copy().items():
            if task_handler(item, task):
                del tasks[item]
        sleep(0.1)


@timed
def iter_filtered_datastores_volumes(client, config, infinibox, datastores, all_volumes):

    log.debug("iter_filtered_datastores_volumes")
    filtered_datastores = list_of_datastores_to_snapshot(
        config, datastores)
    log.debug("list_of_datastores_to_snapshot")
    infi_datastores = iter_filtered_datastores_block_devices(
        client, filtered_datastores)
    log.debug("iter_filtered_datastores_block_devices")
    for datastore, infinidat_devices in infi_datastores:
        for device in infinidat_devices:
            # volume = infinibox.objects.volumes.find(serial=device)
            # Doesn't work anymore with AA - ticket INFINIBOX-38512
            volume = all_volumes.get(device)
            if volume:
                yield datastore, datastore.name, datastore.vm, [volume]


@timed
def get_datastore_infinibox_volumes(client, config, infinibox, datastores, all_volumes):
    return list(iter_filtered_datastores_volumes(client, config, infinibox, datastores, all_volumes))


def get_datastore_snapshot(infinibox_snapshot, datastores_volumes, remote_snapshot):
    for datastore, datastore_name, datastore_vms, volume_list in datastores_volumes:
        if infinibox_snapshot.get_parent() in volume_list:
            return datastore, datastore_name, datastore_vms, infinibox_snapshot, remote_snapshot


def get_vms_infinisnapid_datastore(datastores_snapshot, processed_vms, consistent_only, include_templates):
    datastore, datastore_names, datastore_vms, infinisnap, remote_snap = datastores_snapshot
    vms = get_all_vms_on_datastore_with_single_datastore(
        datastore, processed_vms, consistent_only, include_templates)
    log.debug('Finished get_vms_infinisnapid_datastore')
    return vms, infinisnap, datastore, remote_snap


def get_vms_in_datastore_within_infinisnapid(datastore_volumes, infinibox_snapshot, datastores, vms, consistent_only, remote_snapshot, include_templates):
    datastore_snapshot = get_datastore_snapshot(
        infinibox_snapshot, datastore_volumes, remote_snapshot)
    log.debug("Finished get_datastore_snapshot")
    return get_vms_infinisnapid_datastore(datastore_snapshot, vms, consistent_only, include_templates)


@timed
def parallel_create_vmware_snapshots(client, parallel, vms, basename, guid):
    vmware_snapshots = []

    def create_snap_task_factory(vm_tuple):
        vm, quiesce, memory = vm_tuple
        return initiate_snapshot_for_vm(client, basename, vm, quiesce=quiesce, memory=memory, guid=guid)

    def create_snap_task_handler(vm_tuple, task):
        try:
            vm = vm_tuple[0]
            state, message = get_task_status(task)
            if state == 'success':  # success
                snapshot = wait_for_result(task)
                vmware_snapshots.append((vm, snapshot))
                log.uinfo(
                    'Successfully created snapshot for VM {} ({})', vm.name, snapshot)
            elif state == 'error':
                log.uwarn(
                    'Snapshot creation failed. VM Name: {}, reason: {}', vm.name, message)
        except Exception as e:
            log.uwarn('Unexpected error in create snap {}'.format(
                repr(e)), exc_info=True)
            state = 'error'
        return state in ('success', 'error')

    pipeline_vm_tasks(
        parallel, vms, create_snap_task_factory, create_snap_task_handler)
    return vmware_snapshots


@timed
def parallel_remove_vmware_snapshots(parallel, vmware_snapshots):
    def remove_snap_task_factory(vm_snapshot_pair):
        return initiate_remove_snapshot(vm_snapshot_pair)

    def remove_snap_task_handler(vm_snapshot_pair, task):
        try:
            vm, snapshot = vm_snapshot_pair
            state, message = get_task_status(task)
            if state == 'success':  # success
                log.uinfo(
                    'Successfully deleted snapshot {} for VM {}', snapshot, vm.name)
            elif state == 'error':
                log.uwarn(
                    'Snapshot deletion failed. VM Name: {}, reason: {}', vm.name, message)
        except Exception as e:
            log.uwarn('Unexpected error in remove snap {}'.format(
                repr(e)), exc_info=True)
            state = 'error'
        return state in ('success', 'error')

    pipeline_vm_tasks(
        parallel, vmware_snapshots, remove_snap_task_factory, remove_snap_task_handler)


@timed
def cleanup_vmware_snapshots(vms, client, basename):
    for vm, quiesce, memory in vms:
        try:
            log.debug('Cleanup VMWare snapshots for host {}', vm.name)
            cleanup_infinidat_snapshots(client, basename, vm)
        except Exception as e:
            log.uwarn('Failed to cleanup snapshot {}'.format(repr(e)))


@timed
def do_infinibox_vmware_snapshots(infinibox, config):
    try:
        parallel = config.vcenter.parallel or DEFAULT_PARALLEL
        basename = config.general.snap_basename
        dryrun = config.general.dryrun
        include_templates = config.vcenter.include_templates
        delete = config.general.delete
        consistent_only = config.vcenter.consistent_only
        suppress_metadata = config.logging.suppress

        client = connect_to_vcenter(config)
        infini_datastores = get_all_infinidat_datastores(client)
        all_volumes = get_all_volumes_by_serial(infinibox)
        datastores_to_snapshot = get_datastore_infinibox_volumes(
            client, config, infinibox, infini_datastores, all_volumes)
        if not datastores_to_snapshot:
            log.uabort("Can't find datastores to snapshot. Exiting.",
                       exit_code=errors.EXIT_CODE_VMWARE_NO_DATASTORES_FOUND)
        all_vms = list(iter_all_vms(client))
        log.info(all_vms)
        all_vms_including_powered_off = list(
            iter_all_vms(client, show_powered_off=True, include_templates=config.vcenter.include_templates))
        powertools_info = None
        todo_volumes = 0
        done_volumes = 0
        if not dryrun:
            powertools_info = powertools.get_powertools_info(client)
        datastore_names = [datastore_name for datastore, datastore_name,
                           datastore_vms, infinibox_volumes in datastores_to_snapshot]
        report_missing_vms(config, datastore_names,
                           all_vms_including_powered_off)
        for datastore_volume_tuple in datastores_to_snapshot:
            datastore, datastore_name, datastore_vms, infinibox_volumes = datastore_volume_tuple
            log.uinfo('Datastore to snap: {}', datastore_name)
            guid = uuid.uuid4().hex

            vms_tuple = list(iter_vms(
                config, [datastore_name], all_vms))
            vms = []

            if not vms_tuple:
                if "\n".join(list(config['vms'].keys())) == '!*':
                    log.info(
                        'All VMS excluded. Continuing infinibox snapshots without OS consistency.')
                else:
                    log.uwarn(
                        'No VMs found. Continuing infinibox snapshots without OS consistency.')
                vms_names = []
            else:
                vms_names = [vmname for vmname, vmdatastore,
                             vm in [vm for vm, quiesce, memory in vms_tuple]]
                for vmobj, quiesce, memory in vms_tuple:
                    vmname, vmdatastore, vm = vmobj
                    res = vm, quiesce, memory
                    vms.append(res)
            log.uinfo('VMS to snap: {}', [str(a) for a in vms_names])
            log.debug("infinibox volumes: {}".format(
                [vol for vol in infinibox_volumes]))
            volume_names = [volume.get_name()
                            for vlist in [infinibox_volumes] for volume in vlist]
            todo_volumes += len(volume_names)

            log.uinfo(
                "Infinibox volumes to snap: {}", [str(a) for a in volume_names])
            vmware_snapshots = None
            if not dryrun and not delete:
                vmware_snapshots = parallel_create_vmware_snapshots(
                    client, parallel, list(vms), basename, guid)

            try:
                for infinibox_volume in volume_names:
                    try:
                        done_volumes += 1
                        snapshot_name, metadata_list, remote_snapshot = do_infinibox_snapshot(
                            infinibox, config, infinibox_volume, is_fs=False)
                        if not snapshot_name:
                            done_volumes -= 1

                    except SystemExit as e:
                        done_volumes -= 1
                        if e.code == errors.EXIT_CODE_IBOX_VOLUME_NOT_FOUND:
                            log.uwarn(
                                "Infinibox volume '{}' not found, skipping it.", infinibox_volume)
                            continue
                        else:
                            raise
                    if not dryrun:
                        if powertools_info:
                            if delete:
                                vms_infinisnapid_datastore = None
                            else:
                                vms_infinisnapid_datastore = get_vms_in_datastore_within_infinisnapid(
                                    datastores_to_snapshot, snapshot_name, infini_datastores, vms, consistent_only, remote_snapshot, include_templates)
                                log.debug('About to register the following {}'.format(
                                    vms_infinisnapid_datastore[0][1]))

                            powertools.register_snap_unregister_snaps(
                                powertools_info, vms_infinisnapid_datastore, metadata_list, guid, suppress_metadata)
                            log.debug("Finished register/unregister")
                        else:
                            log.uinfo(
                                'No Host Power Tools For VMWare found. Check appliance installed/powered on.')
            finally:
                if not dryrun:
                    parallel_remove_vmware_snapshots(
                        parallel, vmware_snapshots)
                    cleanup_vmware_snapshots(vms, client, basename)
        return todo_volumes, done_volumes
    finally:
        try:
            client.logout()
        except Exception:
            pass
